from __future__ import print_function
from __future__ import unicode_literals
# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.
#
# unit tests for parse.py
try:
    from unittest import mock
except ImportError:
    import mock

from builtins import zip
from crmsh import parse
import unittest
import shlex
from crmsh.utils import lines2cli
from crmsh.xmlutil import xml_tostring
from lxml import etree


def test_score_to_kind():
    assert parse.score_to_kind("0") == "Optional"
    assert parse.score_to_kind("INFINITY") == "Mandatory"
    assert parse.score_to_kind("200") == "Mandatory"


class MockValidation(parse.Validation):
    def resource_roles(self):
        return ['Master', 'Slave', 'Started']

    def resource_actions(self):
        return ['start', 'stop', 'promote', 'demote']

    def date_ops(self):
        return ['lt', 'gt', 'in_range', 'date_spec']

    def expression_types(self):
        return ['normal', 'string', 'number']

    def rsc_order_kinds(self):
        return ['Mandatory', 'Optional', 'Serialize']

    def op_attributes(self):
        return ['id', 'name', 'interval', 'timeout', 'description',
                'start-delay', 'interval-origin', 'timeout', 'enabled',
                'record-pending', 'role', 'requires', 'on-fail']

    def acl_2_0(self):
        return True


class TestBaseParser(unittest.TestCase):
    def setUp(self):
        self.base = parse.BaseParser()

    def _reset(self, cmd):
        self.base._cmd = shlex.split(cmd)
        self.base._currtok = 0

    @mock.patch('logging.Logger.error')
    def test_err(self, mock_err):
        self._reset('a:b:c:d')

        def runner():
            self.base.match_split()
        self.assertRaises(parse.ParseError, runner)

    @mock.patch('logging.Logger.error')
    def test_idspec(self, mock_error):
        self._reset('$id=foo')
        self.base.match_idspec()
        self.assertEqual(self.base.matched(1), '$id')
        self.assertEqual(self.base.matched(2), 'foo')

        self._reset('$id-ref=foo')
        self.base.match_idspec()
        self.assertEqual(self.base.matched(1), '$id-ref')
        self.assertEqual(self.base.matched(2), 'foo')

        def runner():
            self._reset('id=foo')
            self.base.match_idspec()
        self.assertRaises(parse.ParseError, runner)

    def test_match_split(self):
        self._reset('resource:role')
        a, b = self.base.match_split()
        self.assertEqual(a, 'resource')
        self.assertEqual(b, 'role')

        self._reset('role')
        a, b = self.base.match_split()
        self.assertEqual(a, 'role')
        self.assertEqual(b, None)

    def test_description(self):
        self._reset('description="this is a description"')
        self.assertEqual(self.base.try_match_description(), 'this is a description')

    def test_nvpairs(self):
        self._reset('foo=bar wiz="fizz buzz" bug= bug2=')
        ret = self.base.match_nvpairs()
        self.assertEqual(len(ret), 4)
        retdict = dict([(r.get('name'), r.get('value')) for r in ret])
        self.assertEqual(retdict['foo'], 'bar')
        self.assertEqual(retdict['bug'], '')
        self.assertEqual(retdict['wiz'], 'fizz buzz')


class TestCliParser(unittest.TestCase):
    def setUp(self):
        parse.validator = MockValidation()
        self.comments = []

    def _parse(self, s):
        return parse.parse(s, comments=self.comments)

    @mock.patch('logging.Logger.error')
    def test_node(self, mock_error):
        out = self._parse('node node-1')
        self.assertEqual(out.get('uname'), 'node-1')

        out = self._parse('node $id=testid node-1')
        self.assertEqual(out.get('id'), 'testid')
        self.assertEqual(out.get('uname'), 'node-1')

        out = self._parse('node 1: node-1')
        self.assertEqual(out.get('id'), '1')
        self.assertEqual(out.get('uname'), 'node-1')

        out = self._parse('node testid: node-1')
        self.assertEqual(out.get('id'), 'testid')
        self.assertEqual(out.get('uname'), 'node-1')

        out = self._parse('node $id=testid node-1:remote')
        self.assertEqual(out.get('id'), 'testid')
        self.assertEqual(out.get('uname'), 'node-1')
        self.assertEqual(out.get('type'), 'remote')

        out = self._parse('node node-1:unknown')
        self.assertFalse(out)

        out = self._parse('node node-1 description="foo bar" attributes foo=bar')
        self.assertEqual(out.get('description'), 'foo bar')
        self.assertEqual(['bar'], out.xpath('instance_attributes/nvpair[@name="foo"]/@value'))

        out = self._parse('node node-1 attributes foo=bar utilization wiz=bang')
        self.assertEqual(['bar'], out.xpath('instance_attributes/nvpair[@name="foo"]/@value'))
        self.assertEqual(['bang'], out.xpath('utilization/nvpair[@name="wiz"]/@value'))

    @mock.patch('logging.Logger.error')
    def test_resources(self, mock_error):
        out = self._parse('primitive www ocf:heartbeat:apache op monitor timeout=10s')
        self.assertEqual(out.get('id'), 'www')
        self.assertEqual(out.get('class'), 'ocf')
        self.assertEqual(['monitor'], out.xpath('//op/@name'))

        out = self._parse('rsc_template public_vm ocf:heartbeat:Xen op start timeout=300s op stop timeout=300s op monitor interval=30s timeout=60s op migrate_from timeout=600s op migrate_to timeout=600s')
        self.assertEqual(out.get('id'), 'public_vm')
        self.assertEqual(out.get('class'), 'ocf')
        #print out

        out = self._parse('primitive st stonith:fence_sbd meta target-role=Started requires=nothing op start timeout=60s op monitor interval=60m timeout=60s')
        self.assertEqual(out.get('id'), 'st')

        out2 = self._parse('primitive st stonith:fence_sbd meta target-role=Started requires=nothing op start timeout=60s op monitor interval=60m timeout=60s')
        self.assertEqual(out2.get('id'), 'st')

        self.assertEqual(xml_tostring(out), xml_tostring(out2))

        out = self._parse('primitive st stonith:fence_sbd meta')
        self.assertEqual(out.get('id'), 'st')

        out = self._parse('ms m0 resource params a=b')
        self.assertEqual(out.get('id'), 'm0')
        print(xml_tostring(out))
        self.assertEqual(['resource'], out.xpath('./crmsh-ref/@id'))
        self.assertEqual(['b'], out.xpath('instance_attributes/nvpair[@name="a"]/@value'))

        out2 = self._parse('ms m0 resource a=b')
        self.assertEqual(out.get('id'), 'm0')
        self.assertEqual(xml_tostring(out), xml_tostring(out2))

        out = self._parse('master ma resource meta a=b')
        self.assertEqual(out.get('id'), 'ma')
        self.assertEqual(['resource'], out.xpath('./crmsh-ref/@id'))
        self.assertEqual(['b'], out.xpath('meta_attributes/nvpair[@name="a"]/@value'))

        out = self._parse('clone clone-1 resource meta a=b')
        self.assertEqual(out.get('id'), 'clone-1')
        self.assertEqual(['resource'], out.xpath('./crmsh-ref/@id'))
        self.assertEqual(['b'], out.xpath('meta_attributes/nvpair[@name="a"]/@value'))

        out = self._parse('group group-1 a')
        self.assertEqual(out.get('id'), 'group-1')
        self.assertEqual(len(out), 1)

        out = self._parse('group group-1 a b c')
        self.assertEqual(len(out), 3)

        out = self._parse('group group-1')
        self.assertFalse(out)

        out = self._parse('group group-1 params a=b')
        self.assertEqual(len(out), 1)
        self.assertEqual(['b'], out.xpath('/group/instance_attributes/nvpair[@name="a"]/@value'))

    def test_heartbeat_class(self):
        out = self._parse('primitive p_node-activate heartbeat:node-activate')
        self.assertEqual(out.get('id'), 'p_node-activate')
        self.assertEqual(out.get('class'), 'heartbeat')
        self.assertEqual(out.get('provider'), None)
        self.assertEqual(out.get('type'), 'node-activate')


    def test_nvpair_ref(self):
        out = self._parse('primitive dummy-0 Dummy params @foo')
        self.assertEqual(out.get('id'), 'dummy-0')
        self.assertEqual(out.get('class'), 'ocf')
        self.assertEqual(['foo'], out.xpath('.//nvpair/@id-ref'))

        out = self._parse('primitive dummy-0 Dummy params @fiz')
        self.assertEqual(out.get('id'), 'dummy-0')
        self.assertEqual(out.get('class'), 'ocf')
        self.assertEqual(['fiz'], out.xpath('.//nvpair/@id-ref'))

    @mock.patch('logging.Logger.error')
    def test_location(self, mock_error):
        out = self._parse('location loc-1 resource inf: foo')
        self.assertEqual(out.get('id'), 'loc-1')
        self.assertEqual(out.get('rsc'), 'resource')
        self.assertEqual(out.get('score'), 'INFINITY')
        self.assertEqual(out.get('node'), 'foo')

        out = self._parse('location loc-1 /foo.*/ inf: bar')
        self.assertEqual(out.get('id'), 'loc-1')
        self.assertEqual(out.get('rsc-pattern'), 'foo.*')
        self.assertEqual(out.get('score'), 'INFINITY')
        self.assertEqual(out.get('node'), 'bar')
        #print out

        out = self._parse('location loc-1 // inf: bar')
        self.assertFalse(out)

        out = self._parse('location loc-1 { one ( two three ) four } inf: bar')
        self.assertEqual(out.get('id'), 'loc-1')
        self.assertEqual(['one', 'two', 'three', 'four'], out.xpath('//resource_ref/@id'))
        self.assertEqual(out.get('score'), 'INFINITY')
        self.assertEqual(out.get('node'), 'bar')
        #print out

        out = self._parse('location loc-1 thing rule role=slave -inf: #uname eq madrid')
        self.assertEqual(out.get('id'), 'loc-1')
        self.assertEqual(out.get('rsc'), 'thing')
        self.assertEqual(out.get('score'), None)

        out = self._parse('location l { a:foo b:bar }')
        self.assertFalse(out)

    @mock.patch('logging.Logger.error')
    def test_colocation(self, mock_error):
        out = self._parse('colocation col-1 inf: foo:master ( bar wiz sequential=yes )')
        self.assertEqual(out.get('id'), 'col-1')
        self.assertEqual(['foo', 'bar', 'wiz'], out.xpath('//resource_ref/@id'))
        self.assertEqual([], out.xpath('//resource_set[@name="sequential"]/@value'))

        out = self._parse(
            'colocation col-1 -20: foo:Master ( bar wiz ) ( zip zoo ) node-attribute="fiz"')
        self.assertEqual(out.get('id'), 'col-1')
        self.assertEqual(out.get('score'), '-20')
        self.assertEqual(['foo', 'bar', 'wiz', 'zip', 'zoo'], out.xpath('//resource_ref/@id'))
        self.assertEqual(['fiz'], out.xpath('//@node-attribute'))

        out = self._parse('colocation col-1 0: a:master b')
        self.assertEqual(out.get('id'), 'col-1')

        out = self._parse('colocation col-1 10: ) bar wiz')
        self.assertFalse(out)

        out = self._parse('colocation col-1 10: ( bar wiz')
        self.assertFalse(out)

        out = self._parse('colocation col-1 10: ( bar wiz ]')
        self.assertFalse(out)

    def test_order(self):
        out = self._parse('order o1 Mandatory: [ A B sequential=true ] C')
        print(xml_tostring(out))
        self.assertEqual(['Mandatory'], out.xpath('/rsc_order/@kind'))
        self.assertEqual(2, len(out.xpath('/rsc_order/resource_set')))
        self.assertEqual(['false'], out.xpath('/rsc_order/resource_set/@require-all'))
        self.assertEqual(['A', 'B', 'C'], out.xpath('//resource_ref/@id'))

        out = self._parse('order o1 Mandatory: [ A B sequential=false ] C')
        self.assertEqual(2, len(out.xpath('/rsc_order/resource_set')))
        #self.assertTrue(['require-all', 'false'] in out.resources[0][1])
        #self.assertTrue(['sequential', 'false'] in out.resources[0][1])
        self.assertEqual(out.get('id'), 'o1')

        out = self._parse('order o1 Mandatory: A B C sequential=false')
        self.assertEqual(1, len(out.xpath('/rsc_order/resource_set')))
        #self.assertTrue(['sequential', 'false'] in out.resources[0][1])
        self.assertEqual(out.get('id'), 'o1')

        out = self._parse('order o1 Mandatory: A B C sequential=true')
        self.assertEqual(1, len(out.xpath('/rsc_order/resource_set')))
        #self.assertTrue(['sequential', 'true'] not in out.resources[0][1])
        self.assertEqual(out.get('id'), 'o1')

        out = self._parse('order c_apache_1 Mandatory: apache:start ip_1')
        self.assertEqual(out.get('id'), 'c_apache_1')

        out = self._parse('order c_apache_2 Mandatory: apache:start ip_1 ip_2 ip_3')
        self.assertEqual(2, len(out.xpath('/rsc_order/resource_set')))
        self.assertEqual(out.get('id'), 'c_apache_2')

        out = self._parse('order o1 Serialize: A ( B C )')
        self.assertEqual(2, len(out.xpath('/rsc_order/resource_set')))
        self.assertEqual(out.get('id'), 'o1')

        out = self._parse('order o1 Serialize: A ( B C ) symmetrical=false')
        self.assertEqual(2, len(out.xpath('/rsc_order/resource_set')))
        self.assertEqual(out.get('id'), 'o1')
        self.assertEqual(['false'], out.xpath('//@symmetrical'))

        out = self._parse('order o1 Serialize: A ( B C ) symmetrical=true')
        self.assertEqual(2, len(out.xpath('/rsc_order/resource_set')))
        self.assertEqual(out.get('id'), 'o1')
        self.assertEqual(['true'], out.xpath('//@symmetrical'))

        inp = 'colocation rsc_colocation-master INFINITY: [ vip-master vip-rep sequential=true ] [ msPostgresql:Master sequential=true ]'
        out = self._parse(inp)
        self.assertEqual(2, len(out.xpath('/rsc_colocation/resource_set')))
        self.assertEqual(out.get('id'), 'rsc_colocation-master')

        out = self._parse('order order_2 Mandatory: [ A B ] C')
        self.assertEqual(2, len(out.xpath('/rsc_order/resource_set')))
        self.assertEqual(out.get('id'), 'order_2')
        self.assertEqual(['Mandatory'], out.xpath('/rsc_order/@kind'))
        self.assertEqual(['false'], out.xpath('//resource_set/@sequential'))

        out = self._parse('order order-1 Optional: group1:stop group2:start')
        self.assertEqual(out.get('id'), 'order-1')
        self.assertEqual(['Optional'], out.xpath('/rsc_order/@kind'))
        self.assertEqual(['group1'], out.xpath('/rsc_order/@first'))
        self.assertEqual(['stop'], out.xpath('/rsc_order/@first-action'))
        self.assertEqual(['group2'], out.xpath('/rsc_order/@then'))
        self.assertEqual(['start'], out.xpath('/rsc_order/@then-action'))

    def test_ticket(self):
        out = self._parse('rsc_ticket ticket-A_public-ip ticket-A: public-ip')
        self.assertEqual(out.get('id'), 'ticket-A_public-ip')

        out = self._parse('rsc_ticket ticket-A_bigdb ticket-A: bigdb loss-policy=fence')
        self.assertEqual(out.get('id'), 'ticket-A_bigdb')

        out = self._parse(
            'rsc_ticket ticket-B_storage ticket-B: drbd-a:Master drbd-b:Master')
        self.assertEqual(out.get('id'), 'ticket-B_storage')

    @mock.patch('logging.Logger.error')
    def test_bundle(self, mock_error):
        out = self._parse('bundle httpd docker image=pcmk:httpd replicas=3 network ip-range-start=10.10.10.123 host-netmask=24 port-mapping port=80 storage storage-mapping target-dir=/var/www/html source-dir=/srv/www options=rw primitive httpd-apache')
        self.assertEqual(out.get('id'), 'httpd')
        self.assertEqual(['pcmk:httpd'], out.xpath('/bundle/docker/@image'))
        self.assertEqual(['httpd-apache'], out.xpath('/bundle/crmsh-ref/@id'))

        out = self._parse('bundle httpd docker image=pcmk:httpd primitive httpd-apache apache')
        self.assertFalse(out)

    @mock.patch('logging.Logger.error')
    def test_op(self, mock_error):
        out = self._parse('monitor apache:Master 10s:20s')
        self.assertEqual(out.get('rsc'), 'apache')
        self.assertEqual(out.get('role'), 'Master')
        self.assertEqual(out.get('interval'), '10s')
        self.assertEqual(out.get('timeout'), '20s')

        out = self._parse('monitor apache 60m')
        self.assertEqual(out.get('rsc'), 'apache')
        self.assertEqual(out.get('role'), None)
        self.assertEqual(out.get('interval'), '60m')

        out = self._parse('primitive rsc_dummy1 Dummy op monitor interval=10 OCF_CHECK_LEVEL=10 timeout=60')
        # incorrect ordering of attributes
        self.assertFalse(out)

    @mock.patch('logging.Logger.error')
    def test_acl(self, mock_error):
        out = self._parse('role user-1 error')
        self.assertFalse(out)
        out = self._parse('user user-1 role:user-1')
        self.assertNotEqual(out, False)

        out = self._parse("role bigdb_admin " +
                          "write meta:bigdb:target-role " +
                          "write meta:bigdb:is-managed " +
                          "write location:bigdb " +
                          "read ref:bigdb")
        self.assertEqual(4, len(out))

        # new type of acls

        out = self._parse("acl_target foo a")
        self.assertEqual('acl_target', out.tag)
        self.assertEqual('foo', out.get('id'))
        self.assertEqual(['a'], out.xpath('./role/@id'))

        out = self._parse("acl_target foo a b")
        self.assertEqual('acl_target', out.tag)
        self.assertEqual('foo', out.get('id'))
        self.assertEqual(['a', 'b'], out.xpath('./role/@id'))

        out = self._parse("acl_target foo a b c")
        self.assertEqual('acl_target', out.tag)
        self.assertEqual('foo', out.get('id'))
        self.assertEqual(['a', 'b', 'c'], out.xpath('./role/@id'))
        out = self._parse("acl_group fee a b c")
        self.assertEqual('acl_group', out.tag)
        self.assertEqual('fee', out.get('id'))
        self.assertEqual(['a', 'b', 'c'], out.xpath('./role/@id'))
        out = self._parse('role fum description="test" read a: description="test2" xpath:*[@name=\\"karl\\"]')
        self.assertEqual(['*[@name="karl"]'], out.xpath('/acl_role/acl_permission/@xpath'))

    def test_xml(self):
        out = self._parse('xml <node uname="foo-1"/>')
        self.assertEqual('node', out.tag)
        self.assertEqual('foo-1', out.get('uname'))

    @mock.patch('logging.Logger.error')
    def test_property(self, mock_error):
        out = self._parse('property stonith-enabled=true')
        self.assertEqual(['true'], out.xpath('//nvpair[@name="stonith-enabled"]/@value'))

        # missing score
        out = self._parse('property rule #uname eq node1 stonith-enabled=no')
        self.assertEqual(['INFINITY'], out.xpath('//@score'))

        out = self._parse('property rule 10: #uname eq node1 stonith-enabled=no')
        self.assertEqual(['no'], out.xpath('//nvpair[@name="stonith-enabled"]/@value'))
        self.assertEqual(['node1'], out.xpath('//expression[@attribute="#uname"]/@value'))

        out = self._parse('property rule +inf: date spec years=2014 stonith-enabled=no')
        self.assertEqual(['no'], out.xpath('//nvpair[@name="stonith-enabled"]/@value'))
        self.assertEqual(['2014'], out.xpath('//date_spec/@years'))

        out = self._parse('rsc_defaults failure-timeout=3m')
        self.assertEqual(['3m'], out.xpath('//nvpair[@name="failure-timeout"]/@value'))

        out = self._parse('rsc_defaults foo: failure-timeout=3m')
        self.assertEqual('foo', out[0].get('id'))
        self.assertEqual(['3m'], out.xpath('//nvpair[@name="failure-timeout"]/@value'))

        out = self._parse('rsc_defaults failure-timeout=3m foo:')
        self.assertEqual(False, out)

    def test_empty_property_sets(self):
        out = self._parse('rsc_defaults defaults:')
        self.assertEqual('<rsc_defaults><meta_attributes id="defaults"/></rsc_defaults>',
                         xml_tostring(out))

        out = self._parse('op_defaults defaults:')
        self.assertEqual('<op_defaults><meta_attributes id="defaults"/></op_defaults>',
                         xml_tostring(out))

    def test_fencing(self):
        # num test nodes are 3

        out = self._parse('fencing_topology')
        expect = '<fencing-topology/>'
        self.assertEqual(expect, xml_tostring(out))

        out = self._parse('fencing_topology poison-pill power')
        expect = '<fencing-topology><fencing-level target="ha-one" index="1" devices="poison-pill"/><fencing-level target="ha-one" index="2" devices="power"/><fencing-level target="ha-three" index="1" devices="poison-pill"/><fencing-level target="ha-three" index="2" devices="power"/><fencing-level target="ha-two" index="1" devices="poison-pill"/><fencing-level target="ha-two" index="2" devices="power"/></fencing-topology>'
        self.assertEqual(expect, xml_tostring(out))

        out = self._parse('fencing_topology node-a: poison-pill power node-b: ipmi serial')
        self.assertEqual(4, len(out))

        devs = ['stonith-vbox3-1-off', 'stonith-vbox3-2-off',
                'stonith-vbox3-1-on', 'stonith-vbox3-2-on']
        out = self._parse('fencing_topology vbox4: %s' % ','.join(devs))
        print(xml_tostring(out))
        self.assertEqual(1, len(out))

    def test_fencing_1114(self):
        """
        Test node attribute fence target assignment
        """
        out = self._parse('fencing_topology attr:rack=1 poison-pill power')
        expect = """<fencing-topology><fencing-level index="1" devices="poison-pill" target-attribute="rack" target-value="1"/><fencing-level index="2" devices="power" target-attribute="rack" target-value="1"/></fencing-topology>"""
        self.assertEqual(expect, xml_tostring(out))

        out = self._parse('fencing_topology attr:rack=1 poison-pill,power')
        expect = '<fencing-topology><fencing-level index="1" devices="poison-pill,power" target-attribute="rack" target-value="1"/></fencing-topology>'
        self.assertEqual(expect, xml_tostring(out))

    @mock.patch('logging.Logger.error')
    def test_tag(self, mock_error):
        out = self._parse('tag tag1: one two three')
        self.assertEqual(out.get('id'), 'tag1')
        self.assertEqual(['one', 'two', 'three'], out.xpath('/tag/obj_ref/@id'))

        out = self._parse('tag tag1:')
        self.assertFalse(out)

        out = self._parse('tag tag1:: foo')
        self.assertFalse(out)

        out = self._parse('tag tag1 foo bar')
        self.assertEqual(out.get('id'), 'tag1')
        self.assertEqual(['foo', 'bar'], out.xpath('/tag/obj_ref/@id'))

    def test_alerts(self):
        "Test alerts (1.1.15+)"
        out = self._parse('alert alert1 /tmp/foo.sh to /tmp/bar.log')
        self.assertEqual(out.get('id'), 'alert1')
        self.assertEqual(['/tmp/foo.sh'],
                         out.xpath('/alert/@path'))
        self.assertEqual(['/tmp/bar.log'],
                         out.xpath('/alert/recipient/@value'))

    def test_alerts_brackets(self):
        "Test alerts w/ brackets (1.1.15+)"
        out = self._parse('alert alert2 /tmp/foo.sh to { /tmp/bar.log meta timeout=10s }')
        self.assertEqual(out.get('id'), 'alert2')
        self.assertEqual(['/tmp/foo.sh'],
                         out.xpath('/alert/@path'))
        self.assertEqual(['/tmp/bar.log'],
                         out.xpath('/alert/recipient/@value'))
        self.assertEqual(['10s'],
                         out.xpath('/alert/recipient/meta_attributes/nvpair[@name="timeout"]/@value'))

    def test_alerts_selectors(self):
        "Test alerts w/ selectors (1.1.17+)"
        out = self._parse('alert alert3 /tmp/foo.sh select nodes fencing attributes { standby shutdown } to { /tmp/bar.log meta timeout=10s }')
        self.assertEqual(out.get('id'), 'alert3')
        self.assertEqual(1, len(out.xpath('/alert/select/select_nodes')))
        self.assertEqual(1, len(out.xpath('/alert/select/select_fencing')))
        self.assertEqual(['standby', 'shutdown'],
                         out.xpath('/alert/select/select_attributes/attribute/@name'))


    def _parse_lines(self, lines):
        out = []
        for line in lines2cli(lines):
            if line is not None:
                tmp = self._parse(line.strip())
                self.assertNotEqual(tmp, False)
                if tmp is not None:
                    out.append(tmp)
        return out

    def test_comments(self):
        outp = self._parse_lines('''
        # comment
        node n1
        ''')
        self.assertNotEqual(-1, xml_tostring(outp[0]).find('# comment'))

    def test_uppercase(self):
        outp = self._parse_lines('''
        PRIMITIVE rsc_dummy ocf:heartbeat:Dummy
        MONITOR rsc_dummy 30
        ''')
        #print outp
        self.assertEqual('primitive', outp[0].tag)
        self.assertEqual('op', outp[1].tag)

        outp = self._parse_lines('''
        PRIMITIVE testfs ocf:heartbeat:Filesystem \
          PARAMS directory="/mnt" fstype="ocfs2" device="/dev/sda1"
        CLONE testfs-clone testfs \
          META ordered="true" interleave="true"
        ''')
        #print outp
        self.assertEqual('primitive', outp[0].tag)
        self.assertEqual('clone', outp[1].tag)

        out = self._parse('LOCATION loc-1 resource INF: foo')
        self.assertEqual(out.get('id'), 'loc-1')
        self.assertEqual(out.get('rsc'), 'resource')
        self.assertEqual(out.get('score'), 'INFINITY')
        self.assertEqual(out.get('node'), 'foo')

        out = self._parse('NODE node-1 ATTRIBUTES foo=bar UTILIZATION wiz=bang')
        self.assertEqual('node-1', out.get('uname'))
        self.assertEqual(['bar'], out.xpath('/node/instance_attributes/nvpair[@name="foo"]/@value'))
        self.assertEqual(['bang'], out.xpath('/node/utilization/nvpair[@name="wiz"]/@value'))

        out = self._parse('PRIMITIVE virtual-ip ocf:heartbeat:IPaddr2 PARAMS ip=192.168.122.13 lvs_support=false OP start timeout=20 interval=0 OP stop timeout=20 interval=0 OP monitor interval=10 timeout=20')
        self.assertEqual(['192.168.122.13'], out.xpath('//instance_attributes/nvpair[@name="ip"]/@value'))

        out = self._parse('GROUP web-server virtual-ip apache META target-role=Started')
        self.assertEqual(out.get('id'), 'web-server')

    def test_nvpair_novalue(self):
        inp = """primitive stonith_ipmi-karl stonith:fence_ipmilan \
        params pcmk_host_list=karl verbose action=reboot \
        ipaddr=10.43.242.221 login=root passwd=dummy method=onoff \
        op start interval=0 timeout=60 \
        op stop interval=0 timeout=60 \
        op monitor interval=600 timeout=60 \
        meta target-role=Started"""

        outp = self._parse_lines(inp)
        self.assertEqual(len(outp), 1)
        self.assertEqual('primitive', outp[0].tag)
        # print xml_tostring(outp[0])
        verbose = outp[0].xpath('//nvpair[@name="verbose"]')
        self.assertEqual(len(verbose), 1)
        self.assertTrue('value' not in verbose[0].attrib)

    @mock.patch('logging.Logger.error')
    def test_configs(self, mock_error):
        outp = self._parse_lines('''
        primitive rsc_dummy ocf:heartbeat:Dummy
        monitor rsc_dummy 30
        ''')
        #print outp
        self.assertEqual(2, len(outp))

        outp = self._parse_lines('''
        primitive testfs ocf:heartbeat:Filesystem \
          params directory="/mnt" fstype="ocfs2" device="/dev/sda1"
        clone testfs-clone testfs \
          meta ordered="true" interleave="true"
        ''')
        #print outp
        self.assertEqual(2, len(outp))

        inp = [
            """node node1 attributes mem=16G""",
            """node node2 utilization cpu=4""",
            """primitive d1 ocf:pacemaker:Dummy \
            operations $id=d1-ops \
            op monitor interval=60m \
            op monitor interval=120m OCF_CHECK_LEVEL=10""",
            """monitor d1 60s:30s""",
            """primitive d2 ocf:heartbeat:Delay \
            params mondelay=60 \
            op start timeout=60s \
            op stop timeout=60s""",
            """monitor d2:Started 60s:30s""",
            """group g1 d1 d2""",
            """primitive d3 ocf:pacemaker:Dummy""",
            """clone c d3 \
            meta clone-max=1""",
            """primitive d4 ocf:pacemaker:Dummy""",
            """ms m d4""",
            """primitive s5 ocf:pacemaker:Stateful \
            operations $id-ref=d1-ops""",
            """primitive s6 ocf:pacemaker:Stateful \
            operations $id-ref=d1""",
            """ms m5 s5""",
            """ms m6 s6""",
            """location l1 g1 100: node1""",
            """location l2 c \
            rule $id=l2-rule1 100: #uname eq node1""",
            """location l3 m5 \
            rule inf: #uname eq node1 and pingd gt 0""",
            """location l4 m5 \
            rule -inf: not_defined pingd or pingd lte 0""",
            """location l5 m5 \
            rule -inf: not_defined pingd or pingd lte 0 \
            rule inf: #uname eq node1 and pingd gt 0 \
            rule inf: date lt "2009-05-26" and \
            date in start="2009-05-26" end="2009-07-26" and \
            date in start="2009-05-26" years="2009" and \
            date date_spec years="2009" hours=09-17""",
            """location l6 m5 \
            rule $id-ref=l2-rule1""",
            """location l7 m5 \
            rule $id-ref=l2""",
            """collocation c1 inf: m6 m5""",
            """collocation c2 inf: m5:Master d1:Started""",
            """order o1 Mandatory: m5 m6""",
            """order o2 Optional: d1:start m5:promote""",
            """order o3 Serialize: m5 m6""",
            """order o4 inf: m5 m6""",
            """rsc_ticket ticket-A_m6 ticket-A: m6""",
            """rsc_ticket ticket-B_m6_m5 ticket-B: m6 m5 loss-policy=fence""",
            """rsc_ticket ticket-C_master ticket-C: m6 m5:Master loss-policy=fence""",
            """fencing_topology st st2""",
            """property stonith-enabled=true""",
            """property $id=cpset2 maintenance-mode=true""",
            """rsc_defaults failure-timeout=10m""",
            """op_defaults $id=opsdef2 record-pending=true"""]

        outp = self._parse_lines('\n'.join(inp))
        a = [xml_tostring(x) for x in outp]
        b = [
            '<node uname="node1"><instance_attributes><nvpair name="mem" value="16G"/></instance_attributes></node>',
            '<node uname="node2"><utilization><nvpair name="cpu" value="4"/></utilization></node>',
            '<primitive id="d1" class="ocf" provider="pacemaker" type="Dummy"><operations id="d1-ops"><op name="monitor" interval="60m"/><op name="monitor" interval="120m"><instance_attributes><nvpair name="OCF_CHECK_LEVEL" value="10"/></instance_attributes></op></operations></primitive>',
            '<op name="monitor" rsc="d1" interval="60s" timeout="30s"/>',
            '<primitive id="d2" class="ocf" provider="heartbeat" type="Delay"><instance_attributes><nvpair name="mondelay" value="60"/></instance_attributes><operations><op name="start" timeout="60s" interval="0s"/><op name="stop" timeout="60s" interval="0s"/></operations></primitive>',
            '<op name="monitor" role="Started" rsc="d2" interval="60s" timeout="30s"/>',
            '<group id="g1"><crmsh-ref id="d1"/><crmsh-ref id="d2"/></group>',
            '<primitive id="d3" class="ocf" provider="pacemaker" type="Dummy"/>',
            '<clone id="c"><meta_attributes><nvpair name="clone-max" value="1"/></meta_attributes><crmsh-ref id="d3"/></clone>',
            '<primitive id="d4" class="ocf" provider="pacemaker" type="Dummy"/>',
            '<master id="m"><crmsh-ref id="d4"/></master>',
            '<primitive id="s5" class="ocf" provider="pacemaker" type="Stateful"><operations id-ref="d1-ops"/></primitive>',
            '<primitive id="s6" class="ocf" provider="pacemaker" type="Stateful"><operations id-ref="d1"/></primitive>',
            '<master id="m5"><crmsh-ref id="s5"/></master>',
            '<master id="m6"><crmsh-ref id="s6"/></master>',
            '<rsc_location id="l1" rsc="g1" score="100" node="node1"/>',
            '<rsc_location id="l2" rsc="c"><rule id="l2-rule1" score="100"><expression operation="eq" attribute="#uname" value="node1"/></rule></rsc_location>',
            '<rsc_location id="l3" rsc="m5"><rule score="INFINITY"><expression operation="eq" attribute="#uname" value="node1"/><expression operation="gt" attribute="pingd" value="0"/></rule></rsc_location>',
            '<rsc_location id="l4" rsc="m5"><rule score="-INFINITY" boolean-op="or"><expression operation="not_defined" attribute="pingd"/><expression operation="lte" attribute="pingd" value="0"/></rule></rsc_location>',
            '<rsc_location id="l5" rsc="m5"><rule score="-INFINITY" boolean-op="or"><expression operation="not_defined" attribute="pingd"/><expression operation="lte" attribute="pingd" value="0"/></rule><rule score="INFINITY"><expression operation="eq" attribute="#uname" value="node1"/><expression operation="gt" attribute="pingd" value="0"/></rule><rule score="INFINITY"><date_expression operation="lt" end="2009-05-26"/><date_expression operation="in_range" start="2009-05-26" end="2009-07-26"/><date_expression operation="in_range" start="2009-05-26"><duration years="2009"/></date_expression><date_expression operation="date_spec"><date_spec years="2009" hours="09-17"/></date_expression></rule></rsc_location>',
            '<rsc_location id="l6" rsc="m5"><rule id-ref="l2-rule1"/></rsc_location>',
            '<rsc_location id="l7" rsc="m5"><rule id-ref="l2"/></rsc_location>',
            '<rsc_colocation id="c1" score="INFINITY" rsc="m6" with-rsc="m5"/>',
            '<rsc_colocation id="c2" score="INFINITY" rsc="m5" rsc-role="Master" with-rsc="d1" with-rsc-role="Started"/>',
            '<rsc_order id="o1" kind="Mandatory" first="m5" then="m6"/>',
            '<rsc_order id="o2" kind="Optional" first="d1" first-action="start" then="m5" then-action="promote"/>',
            '<rsc_order id="o3" kind="Serialize" first="m5" then="m6"/>',
            '<rsc_order id="o4" kind="Mandatory" first="m5" then="m6"/>',
            '<rsc_ticket id="ticket-A_m6" ticket="ticket-A" rsc="m6"/>',
            '<rsc_ticket id="ticket-B_m6_m5" ticket="ticket-B" loss-policy="fence"><resource_set><resource_ref id="m6"/><resource_ref id="m5"/></resource_set></rsc_ticket>',
            '<rsc_ticket id="ticket-C_master" ticket="ticket-C" loss-policy="fence"><resource_set><resource_ref id="m6"/></resource_set><resource_set role="Master"><resource_ref id="m5"/></resource_set></rsc_ticket>',
            '<fencing-topology><fencing-level target="ha-one" index="1" devices="st"/><fencing-level target="ha-one" index="2" devices="st2"/><fencing-level target="ha-three" index="1" devices="st"/><fencing-level target="ha-three" index="2" devices="st2"/><fencing-level target="ha-two" index="1" devices="st"/><fencing-level target="ha-two" index="2" devices="st2"/></fencing-topology>',
            '<cluster_property_set><nvpair name="stonith-enabled" value="true"/></cluster_property_set>',
            '<cluster_property_set id="cpset2"><nvpair name="maintenance-mode" value="true"/></cluster_property_set>',
            '<rsc_defaults><meta_attributes><nvpair name="failure-timeout" value="10m"/></meta_attributes></rsc_defaults>',
            '<op_defaults><meta_attributes id="opsdef2"><nvpair name="record-pending" value="true"/></meta_attributes></op_defaults>',
            ]

        for result, expected in zip(a, b):
            self.maxDiff = None
            self.assertEqual(expected, result)

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

# Revised UI structure for crmsh
#
# Goals:
#
# - Modularity
# - Reduced global state
# - Separate static hierarchy from current context
# - Fix completion
# - Implement bash completion
# - Retain all previous functionality
# - Have per-level pre-requirements:
#   def requires(self): <- raise error if prereqs are not met
#   This is so that crmsh can be installed with minimal prereqs,
#   and use cluster sublevel to install all requirements

from . import command
from . import completers as compl
from . import cmd_status
from . import ui_cib
from . import ui_cibstatus
from . import ui_cluster
from . import ui_configure
from . import ui_corosync
from . import ui_history
from . import ui_maintenance
from . import ui_node
from . import ui_options
from . import ui_ra
from . import ui_resource
from . import ui_script
from . import ui_site
from . import ui_sbd


class Root(command.UI):
    """
    Root of the UI hierarchy.
    """

    # name is the user-visible name of this CLI level.
    name = 'root'

    @command.level(ui_cib.CibShadow)
    @command.help('''manage shadow CIBs
A shadow CIB is a regular cluster configuration which is kept in
a file. The CRM and the CRM tools may manage a shadow CIB in the
same way as the live CIB (i.e. the current cluster configuration).
A shadow CIB may be applied to the cluster in one step.
''')
    def do_cib(self):
        pass

    @command.level(ui_cibstatus.CibStatusUI)
    @command.help('''CIB status management and editing
Enter edit and manage the CIB status section level.
''')
    def do_cibstatus(self):
        pass

    @command.level(ui_cluster.Cluster)
    @command.help('''Cluster setup and management
Commands at this level enable low-level cluster configuration
management with HA awareness.
''')
    def do_cluster(self):
        pass

    @command.level(ui_configure.CibConfig)
    @command.help('''CRM cluster configuration
The configuration level.

Note that you can change the working CIB at the cib level. It is
advisable to configure shadow CIBs and then commit them to the
cluster.
''')
    def do_configure(self):
        pass

    @command.level(ui_corosync.Corosync)
    @command.help('''Corosync configuration management
Corosync is the underlying messaging layer for most HA clusters.
This level provides commands for editing and managing the corosync
configuration.
''')
    def do_corosync(self):
        pass

    @command.level(ui_history.History)
    @command.help('''CRM cluster history
The history level.

Examine Pacemaker's history: node and resource events, logs.
''')
    def do_history(self):
        pass

    @command.level(ui_maintenance.Maintenance)
    @command.help('''maintenance
Commands that should only be executed while in
maintenance mode.
''')
    def do_maintenance(self):
        pass

    @command.level(ui_node.NodeMgmt)
    @command.help('''nodes management
A few node related tasks such as node standby are implemented
here.
''')
    def do_node(self):
        pass

    @command.level(ui_options.CliOptions)
    @command.help('''user preferences
Several user preferences are available. Note that it is possible
to save the preferences to a startup file.
''')
    def do_options(self):
        pass

    @command.level(ui_ra.RA)
    @command.help('''resource agents information center
This level contains commands which show various information about
the installed resource agents. It is available both at the top
level and at the `configure` level.
''')
    def do_ra(self):
        pass

    @command.help('''Utility to collect logs and other information
`report` is a utility to collect all information (logs,
configuration files, system information, etc) relevant to
crmsh over the given period of time.
''')
    def do_report(self, context, *args):
        import sys
        from crmsh.report import core
        sys.argv[1:] = args
        core.run()

    @command.level(ui_resource.RscMgmt)
    @command.help('''resources management
Everything related to resources management is available at this
level. Most commands are implemented using the crm_resource(8)
program.
''')
    def do_resource(self):
        pass

    @command.level(ui_sbd.SBD)
    def do_sbd(self):
        pass

    @command.level(ui_script.Script)
    @command.help('''Cluster scripts
Cluster scripts can perform cluster-wide configuration,
validation and management. See the `list` command for
an overview of available scripts.
''')
    def do_script(self):
        pass

    @command.level(ui_site.Site)
    @command.help('''Geo-cluster support
The site level.

Geo-cluster related management.
''')
    def do_site(self):
        pass

    @command.completers(compl.choice(cmd_status.CRM_MON_OPTIONS_MAP))
    @command.help('''show cluster status
Show cluster status. The status is displayed by `crm_mon`. Supply
additional arguments for more information or different format.
See `crm_mon(8)` for more details.

Usage:
...............
status [<option> ...]

option :: bynode | inactive | ops | timing | failcounts
...............
''')
    def do_status(self, context, *args):
        return cmd_status.cmd_status(args)

    @command.help('''Verify cluster state
Performs basic checks for the cluster configuration and
current status, reporting potential issues.

Usage:
.................
verify [scores]
.................
''')
    def do_verify(self, context, *args):
        return cmd_status.cmd_verify(args)


# this will initialize _children for all levels under the root
Root.init_ui(list())


# vim:ts=4:sw=4:et:



#!/usr/bin/python3
# Copyright (C) 2017 Xin Liang <XLiang@suse.com>
# See COPYING for license information.

import argparse
import multiprocessing
import os
import re
import subprocess
import sys
import shutil
import json
import ast
from inspect import getmembers, isfunction
from io import StringIO
from typing import List

import crmsh.sh
import crmsh.report.sh
import crmsh.user_of_host
from crmsh import utils as crmutils, userdir
from crmsh import constants as crmconstants
from crmsh import config, log, tmpfiles, ui_cluster
from crmsh.sh import ShellUtils


logger = log.setup_report_logger(__name__)


class Context:
    """
    Class to set/get essential attributes during the whole crm report process
    """
    def load(self) -> None:
        """
        Load default values
        """
        self.name = "crm_report"
        self.from_time: float = config.report.from_time
        self.to_time: float = utils.now()
        self.no_compress: bool = not config.report.compress
        self.speed_up: bool = config.report.speed_up
        self.extra_log_list: List[str] = config.report.collect_extra_logs
        self.rm_exist_dest: bool = config.report.remove_exist_dest
        self.single: bool= config.report.single_node
        self.sensitive_regex_list: List[str] = []
        self.regex_list: List[str] = "CRIT: ERROR: error: warning: crit:".split()
        self.passwordless_shell_for_nodes = dict()
        self.me = crmutils.this_node()
        self.pe_dir: str
        self.cib_dir: str
        self.pcmk_lib_dir: str
        self.pcmk_exec_dir: str
        self.cores_dir_list: List[str]
        self.trace_dir_list: List[str]
        self.dest: str
        self.dest_dir: str
        self.work_dir: str
        self.node_list: List[str]
        self.ssh_user: str
        self.ssh_option_list: List[str]
        self.no_log_list: List[str]
        self.sanitize: bool
        self.debug: int
        self.compress_prog: str
        self.compress_suffix: str
        self.main_node = self.me

    def __str__(self) -> str:
        return json.dumps({k: v for k, v in self.__dict__.items() if k != 'passwordless_shell_for_nodes'})

    def __setattr__(self, name: str, value) -> None:
        """
        Set the attribute value and perform validations
        """
        if name in ["from_time", "to_time"] and value:
            value = utils.parse_to_timestamp(value)
        if name == "extra_log_list" and value and hasattr(self, "extra_log_list"):
            value = list(set(self.extra_log_list) | set(value))
        super().__setattr__(name, value)

    def __setitem__(self, key: str, value) -> None:
        self.__dict__[key] = value


from crmsh.report import constants, utils, collect


class CapitalizedHelpFormatter(argparse.HelpFormatter):
    def add_usage(self, usage, actions, groups, prefix=None):
        if prefix is None:
            prefix = 'Usage: '
        return super().add_usage(usage.capitalize(), actions, groups, prefix)

    def start_section(self, heading):
        return super().start_section(heading.capitalize())


def add_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
            usage=f"\n{constants.NAME} [options] [dest]",
            add_help=False,
            formatter_class=lambda prog: CapitalizedHelpFormatter(prog, width=80)
            )
    parser.add_argument("-h", "--help", action="store_true", dest="help",
                        help="Show this help message and exit")
    parser.add_argument('-f', dest='from_time', metavar='FROM_TIME',
                        help='Time to start from (default: 12 hours ago), can be specific time or delta time until now')
    parser.add_argument('-t', dest='to_time', metavar='TO_TIME',
                        help='Time to finish at (default: now); can be specific time or delta time until now')
    parser.add_argument('-d', dest='no_compress', action='store_true',
                        help="Don't compress, but leave result in a directory")
    parser.add_argument('-n', dest='node_list', metavar='NODE', action=ui_cluster.CustomAppendAction, default=[],
                        help='Node names for this cluster; this option is additive (use -n a -n b or -n "a b")')
    parser.add_argument('-u', dest='ssh_user', metavar='SSH_USER',
                        help='SSH user to access other nodes')
    parser.add_argument('-X', dest='ssh_option_list', metavar='SSH_OPTION', action=ui_cluster.CustomAppendAction, default=[],
                        help='Extra ssh(1) options; this option is additive')
    parser.add_argument('-E', dest='extra_log_list', metavar='FILE', action=ui_cluster.CustomAppendAction, default=[],
                        help='Extra logs to collect; this option is additive')
    parser.add_argument('-e', dest='no_log_list', metavar='FILE', action=ui_cluster.CustomAppendAction, default=[],
                        help='Don\'t collect these files; this option is additive')
    parser.add_argument('-s', dest='sanitize', action='store_true',
                        help='Replace sensitive info in PE or CIB or pacemaker log files')
    parser.add_argument('-p', dest='sensitive_regex_list', metavar='PATT', action=ui_cluster.CustomAppendAction, default=[],
                        help='Regular expression to match variables containing sensitive data (default: passw.*); this option is additive')
    parser.add_argument('-Q', dest='speed_up', action='store_true',
                        help="The quick mode, which skips producing dot files from PE inputs, verifying installed cluster stack rpms and sanitizing files for sensitive information")
    parser.add_argument('-Z', dest='rm_exist_dest', action='store_true',
                        help='If destination directories exist, remove them instead of exiting')
    parser.add_argument('-S', dest='single', action='store_true',
                        help="Single node operation; don't try to start report collectors on other nodes")
    parser.add_argument('-v', dest='debug', action='count', default=0,
                        help='Increase verbosity')
    parser.add_argument('dest', nargs='?',
                        help="Report name (which may include the path for storing the report), default format is 'crm_report-current_date,' such as 'crm_report-Mon-09-Oct-2023'")

    args = parser.parse_args()
    if args.help:
        print(constants.DESCRIPTION_HELP)
        parser.print_help()
        print(constants.EXTRA_HELP)
        sys.exit(0)

    return args


def push_data(context: Context) -> None:
    """
    Push data from this node
    """
    logger.debug2(f"Pushing data from {context.me}:{context.work_dir} to {context.main_node}")
    cmd = f'cd {context.work_dir}/.. && tar -h -c {context.me}'
    _, out, err = ShellUtils().get_stdout_stderr(cmd, raw=True)
    if out:
        print(f"{constants.COMPRESS_DATA_FLAG}{out}")
    if err:
        raise utils.ReportGenericError(crmutils.to_ascii(err))


def pick_compress_prog(context: Context) -> None:
    """
    Pick the appropriate compression program and its file suffix
    """
    context.compress_prog, context.compress_suffix = pick_first_compress()
    if not context.compress_prog:
        context.compress_prog, context.compress_suffix = "cat", ""


def pick_first_compress():
    compress_prog_suffix_dict = {
        "gzip": ".gz",
        "bzip2": ".bz2",
        "xz": ".xz"
    }
    default_prog = config.report.compress_prog
    default_suffix = compress_prog_suffix_dict.get(default_prog)
    if shutil.which(default_prog) and default_suffix:
        return default_prog, default_suffix
    for cmd, suffix in compress_prog_suffix_dict.items():
        if shutil.which(cmd):
            return cmd, suffix
    logger.warning("Could not find a compression program")
    return None, None


def finalword(context: Context) -> None:
    logger.info(f"The report is saved in {context.dest_path}")
    timespan_str = utils.get_timespan_str(context)
    logger.info(f"Report timespan: {timespan_str}")
    nodes_str = ' '.join(context.node_list)
    logger.info(f"Including nodes: {nodes_str}")
    logger.info("Thank you for taking time to create this report")


def process_results(context: Context) -> None:
    """
    Process report results
    """
    if not context.speed_up:
        utils.do_sanitize(context)
    utils.analyze(context)
    utils.create_description_template(context)

    if context.no_compress:
        shutil.move(context.work_dir, context.dest_dir)
    else:
        cmd_cd_tar = f"(cd {context.work_dir}/.. && tar cf - {context.dest})"
        cmd_compress = f"{context.compress_prog} > {context.dest_dir}/{context.dest}.tar{context.compress_suffix}"
        cmd = f"{cmd_cd_tar}|{cmd_compress}"
        logger.debug2(f"Running: {cmd}")
        crmsh.sh.cluster_shell().get_stdout_or_raise_error(cmd)

    finalword(context)


def collect_logs_and_info(context: Context) -> None:
    """
    Collect logs and information using multiprocessing
    """
	# Make sure not to occupy all CPUs
    pool = multiprocessing.Pool(round(0.8 * multiprocessing.cpu_count()))
    # result here to store AsyncResult object returned from apply_async
    # Then calling get() method will catch exceptions like NameError, AttributeError, etc.
    # Otherwise parent process will not know these exceptions raised
    # Calling get() right after apply_async will be blocked until child process finished, so
    # need to append to a list firstly
    result_list = []
    # Generate function list from collect.py
    for cf in [f for f, _ in getmembers(collect, isfunction) if f.startswith("collect_")]:
        result = pool.apply_async(getattr(collect, cf), (context,))
        result_list.append(result)
    pool.close()
    pool.join()

    for result in result_list:
        try:
            result.get()
        except:
            utils.print_traceback()


def collect_for_nodes(context: Context) -> None:
    """
    Start collectors on each node
    """
    process_list = []
    if len(context.passwordless_shell_for_nodes) == len(context.node_list):
        for node in context.node_list:
            p = multiprocessing.Process(target=start_collector, args=(node, context))
            p.start()
            process_list.append(p)
        for p in process_list:
            p.join()
    else:
        for node in context.node_list:
            start_collector(node, context)


def start_collector(node: str, context: Context) -> None:
    """
    Start collector at specific node
    """
    cmd = f"{constants.BIN_COLLECTOR} '{context}'"

    shell = context.passwordless_shell_for_nodes.get(node)
    if shell is not None:
        ret = shell.subprocess_run_without_input(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    else:
        # This is a last effort when passwordless ssh is not available
        cmd = cmd.replace('"', '\\"')
        ret = _interactive_ssh_run_as_root(node, context.ssh_user, cmd)
    if ret.returncode != 0:
        logger.warning(
            'Failed to run collector on %s: %s: %s',
            node, ret.returncode, crmsh.sh.Utils.decode_str(ret.stderr) if ret.stderr is not None else '',
        )
        return
    elif ret.stderr:
        print(crmsh.sh.Utils.decode_str(ret.stderr), file=sys.stderr)

    archive_data_literal = ""
    for data in ret.stdout.decode('utf-8').split("\n"):
        if data.startswith(constants.COMPRESS_DATA_FLAG):
            # crm report data from collector
            archive_data_literal = data.lstrip(constants.COMPRESS_DATA_FLAG)
        else:
            # INFO log data from collector
            print(data)

    try:
        # Safely evaluate the string representation of a tarball from push_data
        archive_data = ast.literal_eval(archive_data_literal)
    except (SyntaxError, ValueError) as e:
        logger.error(f"Error evaluating data: {e}")
        return

    # Extract the tarball in the specified working directory
    child = subprocess.Popen(
        ['tar', '-x'],
        cwd=context.work_dir,
        stdin=subprocess.PIPE,
    )
    child.stdin.write(archive_data)
    child.stdin.close()
    child.wait()


def process_dest(context: Context) -> None:
    """
    Process destination path and file
    """
    if not context.dest:
        suffix = utils.now(constants.RESULT_TIME_SUFFIX)
        context.dest = f"{context.name}-{suffix}"

    dest_dir = os.path.dirname(context.dest) or "."
    if not os.path.isdir(dest_dir):
        raise utils.ReportGenericError(f"Directory {dest_dir} does not exist")
    context.dest_dir = dest_dir

    dest_file = os.path.basename(context.dest)
    if not crmutils.is_filename_sane(dest_file):
        raise utils.ReportGenericError(f"{dest_file} is invalid file name")

    if context.no_compress and os.path.isdir(context.dest):
        if context.rm_exist_dest:
            shutil.rmtree(context.dest)
        else:
            raise utils.ReportGenericError(f"Destination directory {context.dest} exists, please cleanup or use -Z option")

    context.dest = dest_file
    pick_compress_prog(context)
    if context.no_compress:
        context.dest_path = f"{context.dest_dir}/{context.dest}"
    else:
        context.dest_path = f"{context.dest_dir}/{context.dest}.tar{context.compress_suffix}"


def local_mode(context: Context) -> bool:
    return context.single or (len(context.node_list) == 1 and context.node_list[0] == context.me)


def process_node_list(context: Context) -> None:
    if not local_mode(context) and config.core.no_ssh:
        logger.error('%s', crmconstants.NO_SSH_ERROR_MSG)
        context.single = True
    if context.single:
        context.node_list = [context.me]
    elif not context.node_list:
        context.node_list = crmutils.list_cluster_nodes()
        if not context.node_list:
            raise utils.ReportGenericError("Could not figure out a list of nodes; is this a cluster node?")

    context.node_list = crmutils.get_reachable_node_list(context.node_list)


def process_arguments(context: Context) -> None:
    if context.to_time <= context.from_time:
        raise ValueError("The start time must be before the finish time")
    process_node_list(context)
    process_dest(context)


def setup_workdir(context: Context) -> None:
    """
    Setup working directory where crm report can put all logs into it
    """
    tmpdir = tmpfiles.create_dir()
    if not is_collector():
        context.work_dir = os.path.join(tmpdir, os.path.basename(context.dest))
    else:
        context.work_dir = os.path.join(tmpdir,
                                        os.path.basename(context.dest),
                                        context.me)
    crmutils.mkdirp(context.work_dir)
    logger.debug2(f"Setup work directory in {context.work_dir}")


def load_context(context: Context) -> None:
    """
    Load context attributes from master process
    """
    for key, value in json.loads(sys.argv[2]).items():
        context[key] = value
    context.me = crmutils.this_node()
    adjust_verbosity(context)
    logger.debug2(f"Loading context from collector: {context}")


def find_ssh_user(context: Context) -> None:
    """
    Finds the SSH user for passwordless SSH access to nodes in the context's node_list
    """
    cluster_shell = crmsh.sh.cluster_shell()
    user_of_host = crmsh.user_of_host.UserOfHost.instance()
    for n in context.node_list:
        ret = crmsh.report.sh.Shell.find_shell(cluster_shell, n, context.ssh_user)
        if ret is not None:
            logger.debug("passwordless ssh to %s is OK", n)
            context.passwordless_shell_for_nodes[n] = ret
        else:
            logger.warning("passwordless ssh to node %s does not work", n)
            if not crmutils.can_ask():
                logger.error('Cannot create a report non-interactively. Interactive authentication is required.')
                if userdir.getuser() == 'hacluster':
                    raise ValueError('Passwordless ssh does not work. Run "crm cluster health hawk2 --fix" to set it up.')
                else:
                    raise ValueError('Cannot create a report.')


def load_from_crmsh_config(context: Context) -> None:
    """
    load context attributes from crmsh.config
    """
    config_context_map = {
        "crm_config": "cib_dir",
        "crm_daemon_dir": "pcmk_exec_dir",
        "pe_state_dir": "pe_dir"
    }
    context_str_map = {
        "cib_dir": "CIB",
        "pcmk_exec_dir": "Pacemaker exec",
        "pe_dir": "PE"
    }
    for config_item, context_attr in config_context_map.items():
        value = getattr(config.path, config_item, None)
        if not value or not os.path.isdir(value):
            raise utils.ReportGenericError(f"Cannot find {context_str_map[context_attr]} directory")
        setattr(context, context_attr, value)


def load_context_attributes(context: Context) -> None:
    """
    load context attributes from crmsh.config and corosync.conf
    """
    load_from_crmsh_config(context)

    context.pcmk_lib_dir = os.path.dirname(context.cib_dir)
    context.cores_dir_list = [os.path.join(context.pcmk_lib_dir, "cores")]
    context.cores_dir_list.extend([constants.COROSYNC_LIB] if os.path.isdir(constants.COROSYNC_LIB) else [])


def load_context_trace_dir_list(context: Context) -> None:
    # since the "trace_dir" attribute been removed from cib after untrace
    # need to parse crmsh log file to extract custom trace ra log directory on each node
    log_contents = ""
    cmd = f"grep 'INFO: Trace for .* is written to ' {log.CRMSH_LOG_FILE}*|grep -v 'collect'"
    for node in context.node_list:
        shell = context.passwordless_shell_for_nodes.get(node)
        if shell is not None:
            ret = shell.subprocess_run_without_input(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        else:
            # This is a last effort when passwordless ssh is not available
            cmd = cmd.replace('"', '\\"')
            ret = _interactive_ssh_run_as_root(node, context.ssh_user, cmd)
        if ret.returncode == 0:
            log_contents += ret.stdout.decode('utf-8')
    context.trace_dir_list = list(set(re.findall("written to (.*)/.*", log_contents)))


def _interactive_ssh_run_as_root(host: str, ssh_user: str, cmd: str):
    # cmd MUST be escaped to make sure it works as an arugment of ssh command
    target = f"{ssh_user}@{host}" if ssh_user else host
    logger.info('Creating ssh connection to %s...', target)
    cmd = 'ssh {} {} "{}{}"'.format(
        constants.SSH_OPTS,
        target,
        '' if ssh_user is None or ssh_user == 'root' else 'sudo ',
        cmd,
    )
    shell = crmsh.sh.LocalShell()
    return shell.su_subprocess_run(None, cmd, tty=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


def adjust_verbosity(context: Context) -> None:
    if context.debug > 0:
        config.report.verbosity = context.debug
    elif config.core.debug:
        config.report.verbosity = 1
        context.debug = 1


def parse_arguments(context: Context) -> None:
    """
    Add, parse and process arguments
    """
    args = add_arguments()
    crmutils.check_empty_option_value(args)
    for arg in vars(args):
        value = getattr(args, arg)
        if value or not hasattr(context, arg):
            setattr(context, arg, value)
    adjust_verbosity(context)
    process_arguments(context)


def is_collector() -> bool:
    """
    collector is for collecting logs and data
    """
    return len(sys.argv) > 1 and sys.argv[1] == "__collector"


def run_impl() -> None:
    """
    Major work flow
    """
    ctx = Context()

    if is_collector():
        load_context(ctx)
    else:
        ctx.load()
        parse_arguments(ctx)
        load_context_attributes(ctx)

    setup_workdir(ctx)

    if is_collector():
        collect_logs_and_info(ctx)
        push_data(ctx)
    else:
        find_ssh_user(ctx)
        load_context_trace_dir_list(ctx)
        collect_for_nodes(ctx)
        process_results(ctx)


def run() -> None:
    """
    crm report entry
    """
    try:
        run_impl()
    except UnicodeDecodeError:
        utils.print_traceback()
        sys.exit(1)
    except utils.ReportGenericError as err:
        if str(err):
            logger.error(str(err))
        sys.exit(1)
# vim:ts=4:sw=4:et:



# Copyright (C) 2012 Dejan Muhamedagic <dmuhamedagic@suse.de>
# See COPYING for license information.

import re
from . import config
from .pacemaker import CrmSchema, PacemakerError
from . import log


logger = log.setup_logger(__name__)
PCMK_MIN_SCHEMA_VERSION = 1.0


def is_supported(name):
    """
    Check if the given name is a supported schema name
    A short form is also accepted where the prefix
    pacemaker- is implied.

    Revision: The pacemaker schema version now
    changes too often for a strict check to make sense.
    Lets just check look for schemas we know we don't
    support.
    """
    name = re.match(r'pacemaker-(\d+\.\d+)$', name)
    if name:
        return float(name.group(1)) >= PCMK_MIN_SCHEMA_VERSION
    return False


def get_attrs(schema, name):
    return {
        'a': schema.get_elem_attrs(name, 'a'),  # all
        'r': schema.get_elem_attrs(name, 'r'),  # required
        'o': schema.get_elem_attrs(name, 'o'),  # optional
    }


def get_subs(schema, name):
    return {
        'a': schema.get_sub_elems(name, 'a'),  # all
        'r': schema.get_sub_elems(name, 'r'),  # required
        'o': schema.get_sub_elems(name, 'o'),  # optional
    }


def get_attr_details_d(schema, name):
    # some attributes' names don't repeat, can do a hash
    # (op)
    d = {}
    for attr_obj in schema.get_elem_attr_objs(name):
        attr_name = schema.get_obj_name(attr_obj)
        d[attr_name] = {
            't': schema.get_attr_type(attr_obj),     # type
            'v': schema.get_attr_values(attr_obj),   # values
            'd': schema.get_attr_default(attr_obj),  # default
        }
    return d


def get_attr_details_l(schema, name):
    # some attributes' names repeat, need a list
    # (date_expression)
    l = []
    for attr_obj in schema.get_elem_attr_objs(name):
        l.append({
            'n': schema.get_obj_name(attr_obj),      # name
            't': schema.get_attr_type(attr_obj),     # type
            'v': schema.get_attr_values(attr_obj),   # values
            'd': schema.get_attr_default(attr_obj),  # default
        })
    return l


_cache_funcs = {
    'attr': get_attrs,
    'sub': get_subs,
    'attr_det': get_attr_details_d,
    'attr_det_l': get_attr_details_l,
}


_crm_schema = None
_store = {}


def reset():
    global _store
    _store = {}


def _load_schema(cib):
    return CrmSchema(cib, config.path.crm_dtd_dir)


def init_schema(cib):
    global _crm_schema
    try:
        _crm_schema = _load_schema(cib)
    except PacemakerError as msg:
        logger.error(msg)
    reset()


def test_schema(cib):
    try:
        crm_schema = _load_schema(cib)
    except PacemakerError as msg:
        logger.error(msg)
        return None
    return crm_schema.validate_name


def validate_name():
    if _crm_schema is None:
        return 'pacemaker-2.0'
    return _crm_schema.validate_name


def get(t, name, subset=None):
    if _crm_schema is None:
        return []
    if t not in _store:
        _store[t] = {}
    if name not in _store[t]:
        _store[t][name] = _cache_funcs[t](_crm_schema, name)
    if subset:
        return _store[t][name][subset]
    else:
        return _store[t][name]


def rng_attr_values(el_name, attr_name):
    ""
    try:
        return get('attr_det', el_name)[attr_name]['v']
    except:
        return []


def rng_attr_values_l(el_name, attr_name):
    ""
    l = get('attr_det_l', el_name)
    l2 = []
    for el in l:
        if el['n'] == attr_name:
            l2 += el['v']
    return l2


def rng_xpath(xpath, namespaces=None):
    if _crm_schema is None:
        return []
    return _crm_schema.rng_xpath(xpath, namespaces=namespaces)
# vim:ts=4:sw=4:et:



# Copyright (C) 2019 Xin Liang <XLiang@suse.com>
# See COPYING for license information.
import typing

from crmsh.sh import Utils
from crmsh.prun import prun


Error = prun.PRunError


def parallax_call(nodes, cmd, *, timeout_seconds: int = -1):
    """
    Executes the given command on a set of hosts, collecting the output, and raise exception when error occurs
    nodes:       a set of hosts
    cmd:         command
    timeout_seconds:    the timeout in seconds.
    Returns [(host, (rc, stdout, stdin)), ...], or raises ValueError when any one of the rc != 0
    """
    results = prun.prun({node: cmd for node in nodes}, timeout_seconds=timeout_seconds)
    for node, result in results.items():
        if isinstance(result, prun.PRunError):
            raise ValueError('Failed to run command {} on {}@{}: {}'.format(cmd, result.user, result.host, result))
        elif result.returncode != 0:
            raise ValueError("Failed on {}: {}".format(
                node,
                Utils.decode_str(result.stderr) if result.stderr is not None else None,
            ))
    return [(node, (result.returncode, result.stdout, result.stderr)) for node, result in results.items()]


def parallax_slurp(nodes: typing.Sequence[str], localdir, filename) -> typing.List[typing.Tuple[str, typing.Union[str, Error]]]:
    """
    Copies from the remote node to the local node
    nodes:       a set of hosts
    localdir:    localpath
    filename:    remote filename want to slurp
    Returns [(host, localpath), ...] or raises ValueError when any one of hosts fails.
    """
    results = prun.pfetch_from_remote(nodes, filename, localdir)
    for node, result in results.items():
        if isinstance(result, prun.PRunError):
            raise ValueError("Failed on {}@{}: {}".format(result.user, node, result))
    return [(k, v) for k, v in results.items()]


def parallax_copy(nodes, src, dst, recursive=False, *, timeout_seconds: int = -1):
    """
    Copies from the local node to a set of remote hosts
    nodes:       a set of hosts
    src:         local path
    dst:         remote path
    recursive:   whether to copy directories recursively
    timeout_seconds:    the timeout in seconds.
    Returns None, or raises ValueError when any one of hosts fails.
    """
    results = prun.pcopy_to_remote(src, nodes, dst, recursive, timeout_seconds=timeout_seconds)
    for node, exc in results.items():
        if exc is not None:
            raise ValueError("Failed on {}@{}: {}".format(exc.user, node, exc))


def parallax_run(nodes, cmd):
    """
    Executes the given command on a set of hosts, collecting the output and any error
    nodes:       a set of hosts
    cmd:         command

    Returns [(host, (rc, stdout, stdin)), ...], or raises ValueError when any one of the hosts fails to start running
    the command.
    """
    results = prun.prun({node: cmd for node in nodes})
    for value in results.values():
        if isinstance(value, prun.PRunError):
            raise ValueError('Failed to run command {} on {}@{}: {}'.format(cmd, value.user, value.host, value))
    return {node: (result.returncode, result.stdout, result.stderr) for node, result in results.items()}



# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

# - Base class for UI levels
# - Decorators and other helper functions for the UI
#   Mostly, what these functions do is store extra metadata
#   inside the functions.

import inspect
import re
import ctypes.util
from . import help as help_module
from . import ui_utils
from . import log
from . import constants
from .utils import fuzzy_get


logger = log.setup_logger(__name__)


def name(n):
    '''
    Overrides the name of the command.
    This is useful to handle commands with
    dashes instead of underlines, or commands
    with awkward names (like commands with a
    leading underscore).
    '''
    def inner(fn):
        setattr(fn, '_name', n)
        return fn
    return inner


def alias(*aliases):
    '''
    Adds aliases for the command. The command
    will also be callable using the alias. The
    command name set in the command context will
    reflect the alias used (so the same command can
    behave differently depending on the alias).
    '''
    def inner(fn):
        setattr(fn, '_aliases', aliases)
        return fn
    return inner


def level(level_class):
    '''
    Changes the command into a level movement.
    Calling the command doesn't actually call the
    member function this decorator is applied to, so
    don't put any code in that function.

    This is a bit awkward, but given how decorators work,
    it's the best I could think of.
    '''
    def inner(fn):
        # check signature of given level function
        _check_args(fn, ('self',))

        setattr(fn, '_ui_type', 'level')
        setattr(fn, '_level', level_class)

        def default(arg, val):
            if not hasattr(fn, arg):
                setattr(fn, arg, val)

        default('_aliases', tuple())
        default('_short_help', None)
        default('_long_help', None)
        return fn
    return inner


def help(doc):
    '''
    Use to set a help text for a command or level
    which isn't documented in crm.8.adoc.

    The first line of the doc string will be used as
    the short help, the rest will be used as the full
    help message.
    '''
    doc_split = doc.split('\n', 1)

    def inner(fn):
        setattr(fn, '_short_help', doc_split[0])
        if len(doc_split) > 1:
            setattr(fn, '_long_help', doc_split[1])
        else:
            setattr(fn, '_long_help', '')
        return fn
    return inner


def skill_level(new_level):
    '''
    Use to set the required skill level of a command:

        @command
        @skill_level('administrator')
        def do_rmrf(self, cmd, args):
            ...
    '''
    if isinstance(new_level, str):
        levels = {'operator': 0, 'administrator': 1, 'expert': 2}
        if new_level.lower() not in levels:
            raise ValueError("Unknown skill level: " + new_level)
        new_level = levels[new_level.lower()]

    def inner(fn):
        setattr(fn, '_skill_level', new_level)
        return fn
    return inner


def wait(fn):
    '''
    A command with this decorator will
    force the interactive shell to wait
    for the command to complete.

        @command
        @wait
        def do_bigop(self, cmd, args):
            ...
    '''
    setattr(fn, '_wait', True)
    return fn


def completer(cb):
    '''
    Use to set a tab completer for the command.
    The completer is called for the command, regardless
    of the number of arguments called so far
    '''
    def inner(fn):
        setattr(fn, '_completer', cb)
        return fn
    return inner


def completers(*fns):
    '''
    Use to set a list of positional tab completers for the command.
    Each completer gets as its argument the command line entered so far,
    and returns a list of possible completions.
    '''
    def cfn(args):
        nargs = len(args) - 1
        if nargs == 0:
            return [args[0]]
        if nargs <= len(fns):
            return fns[nargs-1](args)
        return []

    def inner(fn):
        setattr(fn, '_completer', cfn)
        return fn
    return inner


def completers_repeating(*fns):
    '''
    Like completers, but calls the last completer
    for any additional arguments
    '''
    def cfn(args):
        nargs = len(args) - 1
        if nargs == 0:
            return [args[0]]
        if nargs <= len(fns):
            return fns[nargs-1](args)
        return fns[-1](args)

    def inner(fn):
        setattr(fn, '_completer', cfn)
        return fn
    return inner


def _cd_completer(args, context):
    """
    more like bash cd completion
    """
    def current_completions():
        return context.current_level().get_completions()

    def is_sublevel(l):
        return context.current_level().is_sublevel(l)

    def next_completions(token):
        info = context.current_level().get_child(token)
        context.enter_level(info.level)
        return [l for l in current_completions() if is_sublevel(l)]

    def prev_completions():
        return [l for l in context.previous_level().get_completions()
                if context.previous_level().is_sublevel(l)]

    if len(args) == 1 and args[0] == 'cd':
        # complete the 'cd' command self
        return current_completions()
    if len(args) == 2:
        if args[1] in current_completions():
            return [args[1] + '/']
        if args[1] == '..' and context.previous_level():
            return ['../']
        if args[1] == '../' and context.previous_level():
            return [args[1] + l for l in prev_completions()]
        if args[1].endswith("/"):
            return [args[1]+l for l in next_completions(args[1].strip('/'))]
        if re.search(r'\.\./.+', args[1]):
            return ['../' + l for l in prev_completions()]
        if re.search(r'.+/.+', args[1]):
            token = args[1].split('/')[0]
            return [token+'/'+l for l in next_completions(token)]
    if len(args) == 3 and not args[-1]:
        # prevent '..' completion happend many times triggerd by Tab
        return

    ret = []
    if context.previous_level():
        ret += ['..']
    # look out where 'cd' command can enter
    return ret + [l for l in current_completions() if is_sublevel(l)]


def _help_completer(args, context):
    current_level_completions = context.current_level().get_completions()
    if context.current_level().name == 'root':
        return help_module.list_help_topics() + current_level_completions
    return current_level_completions


class UI(object):
    '''
    Base class for all ui levels.
    Things that I need to solve:
      - Error handling
      - Help
      - Completion
    '''

    # Name of level: override this in the subclass.
    name = None

    def requires(self):
        '''
        Returns False if requirements for level are
        not met. Checked before entering the level.
        '''
        return True

    def end_game(self, no_questions_asked=False):
        '''
        Overriding end_game() allows levels to ask
        for confirmation before exiting.
        '''
        pass

    def should_wait(self):
        '''
        A kludge to allow in-transit configuration changes to
        make us wait on transition to finish. Needs to be
        implemented in the level (currently, just configure).
        '''
        return False

    @alias('end', 'back')
    @help('''Go back to previous level
Navigates back in the user interface.
''')
    def do_up(self, context):
        '''
        TODO: Implement full cd navigation. cd ../configure, for example
        Also implement ls to list commands / levels from current location
        '''
        ok = context.up()
        context.save_stack()
        return ok

    @help('''List levels and commands
Lists the available sublevels and commands
at the current level.
''')
    def do_ls(self, context):
        '''
        Shows list of places to go and commands to call
        '''
        out = []
        res = []
        max_width = 16
        if context.previous_level():
            out = ['..']
        out += context.current_level().get_completions()
        for o in sorted(out):
            if o.startswith('-') or o.startswith('_'):
                continue
            if max_width < len(o):
                max_width = len(o)
            res.append(o)

        if max_width >= 16:
            max_width += 2

        colnum = 3
        rownum = len(res) // colnum
        for i in range(rownum):
            for x in res[i::rownum]:
                print("%-0*s" % (max_width, x), end=' ')
            print('')


    @help('''Navigate the level structure
This command works similar to how `cd` works in a regular unix
system shell. `cd ..` returns to the previous level.

If the current level is `resource`, executing `cd ../configure` will
move directly to the `configure` level.

One difference between this command and the usual behavior of `cd`
is that without any argument, this command will go back one level
instead of doing nothing.

Examples:
....
        cd ..
        cd configure
        cd ../configure
        cd configure/ra
....
''')
    @completer(_cd_completer)
    def do_cd(self, context, optarg='..'):
        ok = True
        path = optarg.split('/', 1)
        if len(path) == 1:
            path = path[0]
            if path == '..':
                ok = context.up()
            elif path == '.' or not path:
                return ok
            else:
                info = context.current_level().get_child(path)
                if not info or not info.level:
                    logger.debug("children: %s", self._children)
                    context.fatal_error("%s not found in %s" % (path, context.current_level()))
                context.enter_level(info.level)
        else:
            if not self.do_cd(context, path[0]):
                ok = False
            if not self.do_cd(context, path[1]):
                ok = False
        context.save_stack()
        return True

    @alias('bye', 'exit')
    @help('''Exit the interactive shell
Terminates `crm` immediately. For some levels, `quit` may
ask for confirmation before terminating, if there are
uncommitted changes to the configuration.
''')
    def do_quit(self, context):
        context.quit()

    @alias('?', '-h', '--help')
    @help('''show help (help topics for list of topics)
The help subsystem consists of the command reference and a list
of topics. The former is what you need in order to get the
details regarding a specific command. The latter should help with
concepts and examples.

Examples:
....
        help Introduction
        help quit
....
''')
    @completer(_help_completer)
    def do_help(self, context, *args):
        """usage: help topic|level|command"""
        levels = [ui.name for ui in context.stack]
        levels = levels[1:]     # drop the 1st element "root"
        levels.extend(args)
        h = help_module.help_contextual(levels)
        if h:
            h.paginate()
        context.command_name = ""

    def get_completions(self):
        '''
        return tab completions
        '''
        return [x for x in self._children.keys() if
                x not in self._aliases]

    def get_child(self, child):
        '''
        Returns child info for the given name, or None
        if the child is not found.

        This tries very hard to find a matching child:
        If none is found, a fuzzy matcher is used to
        pick a close match
        '''
        from . import options
        if options.shell_completion:
            return self._children.get(child)
        else:
            return fuzzy_get(self._children, child)

    def is_sublevel(self, child):
        '''
        True if the given name is a sublevel of this level
        '''
        sub = self.get_child(child)
        return sub and sub.type == 'level'

    @classmethod
    def children(cls):
        return cls._children

    @classmethod
    def init_ui(cls, path: list[str]):
        def get_if_command(attr):
            "Return the named attribute if it's a command"
            child = getattr(cls, attr)
            return child if attr.startswith('do_') and inspect.isfunction(child) else None

        def add_aliases(children, info, aliases):
            "Add any aliases for command to child map"
            for alias in info.aliases:
                aliases.append(alias)
                children[alias] = info

        def add_help(path: list[str], info):
            "Add static help to the help system"
            if info.short_help:
                entry = help_module.HelpEntry(info.short_help, info.long_help, generated=True)
                help_module.add_help(path + [info.name], entry)
            elif info.type == 'command':
                entry = help_module.HelpEntry(
                    'Help for command ' + info.name,
                    'Note: This command is not documented.\n' +
                    'Usage: %s %s' % (info.name,
                                      ui_utils.pretty_arguments(info.function, nskip=2)),
                    generated=True)
                help_module.add_help(path + [info.name], entry)
            elif info.type == 'level':
                entry = help_module.HelpEntry('Help for level ' + info.name,
                                              'Note: This level is not documented.\n',
                                              generated=True)
                help_module.add_help(path + [info.name], entry)

        children = {}
        aliases = []
        for child_name in dir(cls):
            if child_name == 'do_up' and re.search("ui_root.Root", str(cls)):
                continue
            child = get_if_command(child_name)
            if child:
                info = ChildInfo(child, cls)
                if info.type == 'command' and not is_valid_command_function(info.function):
                    raise ValueError("Invalid command function: %s.%s" %
                                     (cls.__name__, info.function.__name__))
                elif info.type == 'level' and info.level:
                    info.children = info.level.init_ui(path + [info.name])
                children[info.name] = info
                add_aliases(children, info, aliases)
                add_help(path, info)
        cls._children = children
        cls._aliases = aliases
        return children


def make_name(new_name):
    '''
    Generate command name from command function name.
    '''
    if new_name.startswith('do_'):
        return new_name[3:]
    return new_name


class ChildInfo(object):
    '''
    Declares the given method a command method.
    Sets extra attributes in the function itself,
    which are picked up by the UILevel class and used
    to generate ChildInfo data.

    The given method is expected to take a first parameter
    (after self) which is a UI context, which holds information
    about where the user came from when calling the command, controls
    for manipulating the current level (up(), quit(), etc),
    the name used when calling the command, error reporting and warning
    methods.

    The rest of the parameters are the actual arguments to the method. These
    are tokenized using shlex and then matched to the actual arguments of the
    method.

    Information about a child node in the hierarchy:
    A node is either a level or a command.
    '''
    def __init__(self, fn, parent):
        def maybe(attr, default):
            if hasattr(fn, attr):
                return getattr(fn, attr)
            return default

        self.function = fn
        self.name = maybe('_name', make_name(fn.__name__))
        self.type = maybe('_ui_type', 'command')
        self.aliases = maybe('_aliases', tuple())
        self.short_help = maybe('_short_help', None)
        self.long_help = maybe('_long_help', None)
        self.skill_level = maybe('_skill_level', 0)
        self.wait = maybe('_wait', False)
        self.level = maybe('_level', None)
        self.completer = maybe('_completer', None)
        self.parent = parent
        self.children = {}

    def complete(self, context, args):
        '''
        Execute the completer for this command with the given arguments.
        The completer mostly completes based on argument position, but
        some commands are context sensitive...
        - make args[0] be name of command
        '''
        ret = []
        if self.completer is not None:
            disable_custom_sort_order()
            specs = inspect.getfullargspec(self.completer)
            if 'context' in specs.args:
                ret = self.completer([self.name] + args, context)
            else:
                ret = self.completer([self.name] + args)
            ret = ret if already_used_custom_sort_order() else sorted(ret)
        return ret

    def __repr__(self):
        return "%s:%s (%s)" % (self.type, self.name, self.short_help)


def is_valid_command_function(fn):
    '''
    Returns True if fn is a valid command function:
    named do_xxx, takes (self, context) as the first two parameters
    '''
    specs = inspect.getfullargspec(fn)
    return len(specs.args) >= 2 and specs.args[0] == 'self' and specs.args[1] == 'context'


def _check_args(fn, expected):
    argnames = fn.__code__.co_varnames[:fn.__code__.co_argcount]
    if argnames != expected:
        raise ValueError(fn.__name__ +
                         ": Expected method with signature " + repr(expected))


def enable_custom_sort_order():
    if sort_completion_inst is not None:
        sort_completion_inst.value = 0


def already_used_custom_sort_order():
    return sort_completion_inst is not None and sort_completion_inst.value == 0


def disable_custom_sort_order():
    if already_used_custom_sort_order():
        # Restore the original value of rl_sort_completion_matches
        # before calling the completer again
        sort_completion_inst.value = orig_sort_completion_value


readline_path = ctypes.util.find_library('readline')
sort_completion_inst = None
if readline_path:
    # Inspired by
    # https://stackoverflow.com/questions/31229708/python-gnu-readline-bindings-keep-my-sort-order
    # Python readline module sort the completion result by
    # alphabetical order by default. To archive the custom
    # sort order, need to disable the sort_completion_matches
    rl = ctypes.cdll.LoadLibrary(readline_path)
    sort_completion_inst = ctypes.c_ulong.in_dll(rl, 'rl_sort_completion_matches')
    orig_sort_completion_value = sort_completion_inst.value



# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.
'''
Holds user-configurable options.
'''

import os
import re
import sys
import configparser
from contextlib import contextmanager
from typing import List
from . import userdir


@contextmanager
def _disable_exception_traceback():
    """
    All traceback information is suppressed and only the exception type and value are printed
    """
    default_value = getattr(sys, "tracebacklimit", 1000)  # `1000` is a Python's default value
    sys.tracebacklimit = 0
    yield
    sys.tracebacklimit = default_value  # revert changes


def configure_libdir():
    '''
    sysconfig is only available in 2.7 and above
    MULTIARCH is a debian specific configuration variable
    '''
    dirs = ('/usr/lib64', '/usr/libexec', '/usr/lib',
            '/usr/local/lib64', '/usr/local/libexec', '/usr/local/lib')
    try:
        import sysconfig
        multiarch = sysconfig.get_config_var('MULTIARCH')
        if multiarch:
            dirs += ('/usr/lib/%s' % multiarch,
                     '/usr/local/lib/%s' % multiarch)
    except ImportError:
        pass
    return dirs


_SYSTEMWIDE = '/etc/crm/crm.conf'
_PERUSER = os.getenv("CRM_CONFIG_FILE") or os.path.join(userdir.CONFIG_HOME, 'crm.conf')

_PATHLIST = {
    'datadir': ('/usr/share', '/usr/local/share', '/opt'),
    'cachedir': ('/var/cache', '/opt/cache'),
    'libdir': configure_libdir(),
    'varlib': ('/var/lib', '/opt/var/lib'),
    'wwwdir': ('/srv/www', '/var/www')
}


def make_path(path):
    """input: path containing %(?)s-statements
    output: path with no such statements"""
    m = re.match(r'\%\(([^\)]+)\)(.+)', path)
    if m:
        t = m.group(1)
        for dd in _PATHLIST[t]:
            if os.path.isdir(path % {t: dd}):
                return path % {t: dd}
        return path % {t: _PATHLIST[t][0]}
    return path


def find_pacemaker_daemons():
    '''
    Search for the pacemaker daemon location by trying to find
    where the daemons are. The control daemon is either
    pacemaker-controld (2.0+) or crmd depending on the version.
    '''
    candidate_dirs = ('{}/pacemaker'.format(d) for d in configure_libdir())
    for d in candidate_dirs:
        daemon = '{}/pacemaker-controld'.format(d)
        if os.path.exists(daemon):
            return d
        daemon = '{}/crmd'.format(d)
        if os.path.exists(daemon):
            return d
    return '/usr/lib/pacemaker'


# opt_ classes
# members: default, completions, validate()

class opt_program(object):
    def __init__(self, envvar, proglist):
        self.default = ''
        if envvar and os.getenv(envvar):
            self.default = os.getenv(envvar)
        else:
            for prog in proglist:
                p = self._find_program(prog)
                if p is not None:
                    self.default = p
                    break
        self.completions = proglist

    def _find_program(self, prog):
        """Is this program available?"""
        paths = os.getenv("PATH").split(os.pathsep)
        paths.extend(['/usr/bin', '/usr/sbin', '/bin', '/sbin'])
        if prog.startswith('/'):
            filename = make_path(prog)
            if os.path.isfile(filename) and os.access(filename, os.X_OK):
                return filename
        elif prog.startswith('%'):
            prog = make_path(prog)
            for p in paths:
                filename = os.path.join(p, prog)
                if os.path.isfile(filename) and os.access(filename, os.X_OK):
                    return filename
        else:
            for p in paths:
                filename = make_path(os.path.join(p, prog))
                if os.path.isfile(filename) and os.access(filename, os.X_OK):
                    return filename
        return None

    def validate(self, prog):
        if self._find_program(prog) is None:
            raise ValueError("%s does not exist or is not a program" % prog)

    def get(self, value):
        if value.startswith('$'):
            return os.getenv(value[1:])
        elif value.startswith('\\$'):
            return value[1:]
        return value


class opt_string(object):
    def __init__(self, value):
        self.default = value
        self.completions = ()

    def validate(self, val):
        return True

    def get(self, value):
        return value


class opt_choice(object):
    def __init__(self, dflt, choices):
        self.default = dflt
        self.completions = choices

    def validate(self, val):
        if val not in self.completions:
            raise ValueError("%s not in %s" % (val, ', '.join(self.completions)))

    def get(self, value):
        return value


class opt_multichoice(object):
    def __init__(self, dflt, choices):
        self.default = dflt
        self.completions = choices

    def validate(self, val):
        vals = [x.strip() for x in val.split(',')]
        for otype in vals:
            if otype not in self.completions:
                raise ValueError("%s not in %s" % (val, ', '.join(self.completions)))

    def get(self, value):
        return value


class opt_boolean(object):
    def __init__(self, dflt):
        self.default = dflt
        self.completions = ('yes', 'true', 'on', '1', 'no', 'false', 'off', '0')

    def validate(self, val):
        if val is True:
            val = 'true'
        elif val is False:
            val = 'false'
        val = val.lower()
        if val not in self.completions:
            raise ValueError("Not a boolean: %s (try one of: %s)" % (
                val, ', '.join(self.completions)))

    def get(self, value):
        return value.lower() in ('yes', 'true', 'on', '1')


class opt_dir(object):
    def __init__(self, path):
        self.default = make_path(path)
        self.completions = []

    def validate(self, val):
        if not os.path.isdir(val):
            raise ValueError("Directory not found: %s" % (val))

    def get(self, value):
        return value


class opt_color(object):
    def __init__(self, val):
        self.default = val
        self.completions = ('black', 'blue', 'green', 'cyan',
                            'red', 'magenta', 'yellow', 'white',
                            'bold', 'blink', 'dim', 'reverse',
                            'underline', 'normal')

    def validate(self, val):
        for v in val.split(' '):
            if v not in self.completions:
                raise ValueError('Invalid color ' + val)

    def get(self, value):
        return [s.rstrip(',') for s in value.split(' ')] or ['normal']


class opt_list(object):
    def __init__(self, deflist):
        self.default = ' '.join(deflist)
        self.completions = deflist

    def validate(self, val):
        pass

    def get(self, value):
        return [s.rstrip(',') for s in value.split(' ')]


DEFAULTS = {
    'core': {
        'editor': opt_program('EDITOR', ('vim', 'vi', 'emacs', 'nano')),
        'pager': opt_program('PAGER', ('less', 'more', 'pg')),
        'user': opt_string(''),
        'hosts': opt_list([]), # 'alice@host1, bob@host2'
        'no_generating_ssh_key': opt_boolean('no'), # deprecated
        'skill_level': opt_choice('expert', ('operator', 'administrator', 'expert')),
        'sort_elements': opt_boolean('yes'),
        'check_frequency': opt_choice('always', ('always', 'on-verify', 'never')),
        'check_mode': opt_choice('strict', ('strict', 'relaxed')),
        'wait': opt_boolean('no'),
        'add_quotes': opt_boolean('yes'),
        'manage_children': opt_choice('ask', ('ask', 'never', 'always')),
        'force': opt_boolean('no'),
        'debug': opt_boolean('no'),
        'ptest': opt_program('', ('ptest', 'crm_simulate')),
        'dotty': opt_program('', ('dotty',)),
        'dot': opt_program('', ('dot',)),
        'ignore_missing_metadata': opt_boolean('no'),
        'report_tool_options': opt_string(''),
        'lock_timeout': opt_string('120'),
        'no_ssh': opt_boolean('no'),
        'has_ra_advised_op': opt_boolean('no'),
        'has_fa_advised_op': opt_boolean('no'),
        'obscure_pattern': opt_string('passw*')
    },
    'path': {
        'sharedir': opt_dir('%(datadir)s/crmsh'),
        'cache': opt_dir('%(cachedir)s/crm'),
        'crm_config': opt_dir('%(varlib)s/pacemaker/cib'),
        'crm_daemon_dir': opt_dir(find_pacemaker_daemons()),
        'crm_daemon_user': opt_string('hacluster'),
        'ocf_root': opt_dir('%(libdir)s/ocf'),
        'crm_dtd_dir': opt_dir('%(datadir)s/pacemaker'),
        'pe_state_dir': opt_dir('%(varlib)s/pacemaker/pengine'),
        'heartbeat_dir': opt_dir('%(varlib)s/heartbeat'),
        'hawk_wizards': opt_dir('%(wwwdir)s/hawk/config/wizard'),
    },
    'color': {
        'style': opt_multichoice('color', ('plain', 'color-always', 'color', 'uppercase')),
        'error': opt_color('red bold'),
        'ok': opt_color('green bold'),
        'warn': opt_color('yellow bold'),
        'info': opt_color('cyan'),
        'help_keyword': opt_color('blue bold underline'),
        'help_header': opt_color('normal bold'),
        'help_topic': opt_color('yellow bold'),
        'help_block': opt_color('cyan'),
        'keyword': opt_color('yellow'),
        'identifier': opt_color('normal'),
        'attr_name': opt_color('cyan'),
        'attr_value': opt_color('red'),
        'resource_reference': opt_color('green'),
        'id_reference': opt_color('green'),
        'score': opt_color('magenta'),
        'ticket': opt_color('magenta'),
    },
    'report': {
        'from_time': opt_string('-12H'),
        'compress': opt_boolean('yes'),
        'compress_prog': opt_string('gzip'),
        'speed_up': opt_boolean('no'),
        'collect_extra_logs': opt_list([
            '/var/log/messages',
            '/var/log/crmsh/crmsh.log',
            '/etc/crm/profiles.yml',
            '/etc/crm/crm.conf',
            '~/.config/crm/crm.conf'
        ]),
        'remove_exist_dest': opt_boolean('no'),
        'single_node': opt_boolean('no'),
        'sanitize_rule': opt_string('passw.*'),
        'verbosity': opt_string('0')
    }
}

_parser = None


def _stringify(val):
    if val is True:
        return 'true'
    elif val is False:
        return 'false'
    elif isinstance(val, str):
        return val
    else:
        return str(val)


class _Configuration(object):
    def __init__(self):
        self._defaults = None
        self._systemwide = None
        self._user = None

    def _safe_read(self, config_parser_inst, file_list):
        """
        Try to handle configparser.MissingSectionHeaderError while reading
        """
        try:
            config_parser_inst.read(file_list)
        except configparser.MissingSectionHeaderError:
            with _disable_exception_traceback():
                raise

    def load(self):
        self._defaults = configparser.ConfigParser()
        for section, keys in DEFAULTS.items():
            self._defaults.add_section(section)
            for key, opt in keys.items():
                self._defaults.set(section, key, opt.default)

        if os.path.isfile(_SYSTEMWIDE):
            self._systemwide = configparser.ConfigParser()
            self._safe_read(self._systemwide, [_SYSTEMWIDE])
        # for backwards compatibility with <=2.1.1 due to ridiculous bug
        elif os.path.isfile("/etc/crm/crmsh.conf"):
            self._systemwide = configparser.ConfigParser()
            self._safe_read(self._systemwide, ["/etc/crm/crmsh.conf"])
        if os.path.isfile(_PERUSER):
            self._user = configparser.ConfigParser()
            self._safe_read(self._user, [_PERUSER])

    def save(self):
        if self._user:
            if not os.path.isdir(os.path.dirname(_PERUSER)):
                os.makedirs(os.path.dirname(_PERUSER))
            fp = open(_PERUSER, 'w')
            self._user.write(fp)
            fp.close()

    def get_impl(self, section, name):
        try:
            if self._user and self._user.has_option(section, name):
                return self._user.get(section, name) or ''
            if self._systemwide and self._systemwide.has_option(section, name):
                return self._systemwide.get(section, name) or ''
            return self._defaults.get(section, name) or ''
        except configparser.NoOptionError as e:
            raise ValueError(e)

    def get(self, section, name, raw=False):
        if raw:
            return self.get_impl(section, name)
        return DEFAULTS[section][name].get(self.get_impl(section, name))

    def set(self, section, name, value):
        if section not in ('core', 'path', 'color', 'report'):
            raise ValueError("Setting invalid section " + str(section))
        if not self._defaults.has_option(section, name):
            raise ValueError("Setting invalid option %s.%s" % (section, name))
        DEFAULTS[section][name].validate(value)
        if self._user is None:
            self._user = configparser.ConfigParser()
        if not self._user.has_section(section):
            self._user.add_section(section)
        self._user.set(section, name, _stringify(value))

    def items(self, section):
        return [(k, self.get(section, k)) for k, _ in self._defaults.items(section)]

    def configured_keys(self, section):
        ret = []
        if self._systemwide and self._systemwide.has_section(section):
            ret += self._systemwide.options(section)
        if self._user and self._user.has_section(section):
            ret += self._user.options(section)
        return list(set(ret))

    def reset(self):
        '''reset to what is on disk'''
        self._user = configparser.ConfigParser()
        self._user.read([_PERUSER])


_configuration = _Configuration()


class _Section(object):
    def __init__(self, section):
        object.__setattr__(self, 'section', section)

    def __getattr__(self, name):
        return _configuration.get(self.section, name)

    def __setattr__(self, name, value):
        _configuration.set(self.section, name, value)

    def items(self):
        return _configuration.items(self.section)


def load():
    _configuration.load()

    os.environ["OCF_ROOT"] = _configuration.get('path', 'ocf_root')


def save():
    '''
    Only save options that are not default
    '''
    _configuration.save()


def set_option(section, option, value):
    if not isinstance(value, List):
        _configuration.set(section, option, value)
        return
    string = ""
    first = True
    for item in value:
        if first:
            first = False
        else:
            string += ", "
        string += str(item)
    _configuration.set(section, option, string)


def get_option(section, option, raw=False):
    '''
    Return the given option.
    If raw is True, return the configured value.
    Example: for a boolean, returns "yes", not True
    '''
    return _configuration.get(section, option, raw=raw)


def get_all_options():
    '''Returns a list of all configurable options'''
    ret = []
    for sname, section in DEFAULTS.items():
        ret += ['%s.%s' % (sname, option) for option in list(section.keys())]
    return sorted(ret)


def get_configured_options():
    '''Returns a list of all options that have a non-default value'''
    ret = []
    for sname in DEFAULTS:
        for key in _configuration.configured_keys(sname):
            ret.append('%s.%s' % (sname, key))
    return ret


def complete(section, option):
    s = DEFAULTS.get(section)
    if not s:
        return []
    o = s.get(option)
    if not o:
        return []
    return o.completions


def has_user_config():
    return os.path.isfile(_PERUSER)


def reset():
    _configuration.reset()


load()
core = _Section('core')
path = _Section('path')
color = _Section('color')
report = _Section('report')


def load_version():
    version = 'dev'
    versioninfo_file = os.path.join(path.sharedir, 'version')
    if os.path.isfile(versioninfo_file):
        with open(versioninfo_file) as f:
            version = f.readline().strip() or version
    return version


VERSION = load_version()
CRM_VERSION = str(VERSION)



import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))

try:
    from unittest import mock, TestCase
except ImportError:
    import mock
import logging

from crmsh.crash_test import utils, main, config


class TestMyLoggingFormatter(TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Global setUp.
        """

    def setUp(self):
        """
        Test setUp.
        """
        self.fence_info_inst = utils.FenceInfo()

    def tearDown(self):
        """
        Test tearDown.
        """

    @classmethod
    def tearDownClass(cls):
        """
        Global tearDown.
        """


class TestFenceInfo(TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Global setUp.
        """

    def setUp(self):
        """
        Test setUp.
        """
        self.fence_info_inst = utils.FenceInfo()

    def tearDown(self):
        """
        Test tearDown.
        """

    @classmethod
    def tearDownClass(cls):
        """
        Global tearDown.
        """

    @mock.patch('crmsh.crash_test.utils.crmshutils.get_property')
    def test_fence_enabled_false(self, mock_get_property):
        mock_get_property.return_value = None
        res = self.fence_info_inst.fence_enabled
        self.assertEqual(res, False)
        mock_get_property.assert_called_once_with("stonith-enabled")

    @mock.patch('crmsh.crash_test.utils.crmshutils.get_property')
    def test_fence_enabled_true(self, mock_get_property):
        mock_get_property.return_value = "True"
        res = self.fence_info_inst.fence_enabled
        self.assertEqual(res, True)
        mock_get_property.assert_called_once_with("stonith-enabled")

    @mock.patch('crmsh.crash_test.utils.msg_error')
    @mock.patch('crmsh.ra.get_property_options')
    @mock.patch('crmsh.crash_test.utils.crmshutils.get_property')
    def test_fence_action_none(self, mock_get_property, mock_get_property_options, mock_error):
        mock_get_property_options.return_value = ["off", "reboot"]
        mock_get_property.return_value = None
        res = self.fence_info_inst.fence_action
        self.assertEqual(res, None)
        mock_get_property.assert_called_once_with("stonith-action")
        mock_error.assert_called_once_with('Cluster property "stonith-action" should be off|reboot')

    @mock.patch('crmsh.ra.get_property_options')
    @mock.patch('crmsh.crash_test.utils.crmshutils.get_property')
    def test_fence_action(self, mock_get_property, mock_get_property_options):
        mock_get_property_options.return_value = ["off", "reboot"]
        mock_get_property.return_value = "reboot"
        res = self.fence_info_inst.fence_action
        self.assertEqual(res, "reboot")
        mock_get_property.assert_called_once_with("stonith-action")

    @mock.patch('crmsh.crash_test.utils.crmshutils.get_property')
    def test_fence_timeout(self, mock_get_property):
        mock_get_property.return_value = "60s"
        res = self.fence_info_inst.fence_timeout
        self.assertEqual(res, "60")
        mock_get_property.assert_called_once_with("stonith-timeout")

    @mock.patch('crmsh.crash_test.utils.crmshutils.get_property')
    def test_fence_timeout_default(self, mock_get_property):
        mock_get_property.return_value = None
        res = self.fence_info_inst.fence_timeout
        self.assertEqual(res, config.FENCE_TIMEOUT)
        mock_get_property.assert_called_once_with("stonith-timeout")


class TestUtils(TestCase):
    '''
    Unitary tests for crash_test/utils.py
    '''

    @classmethod
    def setUpClass(cls):
        """
        Global setUp.
        """

    def setUp(self):
        """
        Test setUp.
        """

    def tearDown(self):
        """
        Test tearDown.
        """

    @classmethod
    def tearDownClass(cls):
        """
        Global tearDown.
        """

    @mock.patch('crmsh.crash_test.utils.datetime')
    def test_now(self, mock_datetime):
        mock_now = mock.Mock()
        mock_datetime.now.return_value = mock_now
        mock_now.strftime.return_value = "2019/07/05 14:44:55"

        result = utils.now()

        self.assertEqual(result, "2019/07/05 14:44:55")
        mock_datetime.now.assert_called_once_with()
        mock_now.strftime.assert_called_once_with("%Y/%m/%d %H:%M:%S")

    @mock.patch('crmsh.crash_test.utils.get_handler')
    def test_manage_handler(self, mock_get_handler):
        mock_get_handler.return_value = "handler"
        utils.logger = mock.Mock()
        utils.logger.removeHandler = mock.Mock()
        utils.logger.addHandler = mock.Mock()

        with utils.manage_handler("type1", keep=False):
            pass

        mock_get_handler.assert_called_once_with(utils.logger, "type1")
        utils.logger.removeHandler.assert_called_once_with("handler")
        utils.logger.addHandler.assert_called_once_with("handler")

    @mock.patch('crmsh.crash_test.utils.manage_handler')
    def test_msg_raw(self, mock_handler):
        utils.logger = mock.Mock()
        utils.logger.log = mock.Mock()
        utils.msg_raw("level1", "msg1")
        mock_handler.assert_called_once_with("console", True)
        utils.logger.log.assert_called_once_with("level1", "msg1")

    @mock.patch('crmsh.crash_test.utils.msg_raw')
    def test_msg_info(self, mock_raw):
        utils.msg_info("msg1")
        mock_raw.assert_called_once_with(logging.INFO, "msg1", True)

    @mock.patch('crmsh.crash_test.utils.msg_raw')
    def test_msg_warn(self, mock_raw):
        utils.msg_warn("msg1")
        mock_raw.assert_called_once_with(logging.WARNING, "msg1", True)

    @mock.patch('crmsh.crash_test.utils.msg_raw')
    def test_msg_error(self, mock_raw):
        utils.msg_error("msg1")
        mock_raw.assert_called_once_with(logging.ERROR, "msg1", True)

    @mock.patch('os.fsync')
    @mock.patch('json.dumps')
    @mock.patch('builtins.open', create=True)
    def test_json_dumps(self, mock_open_file, mock_dumps, mock_fsync):
        main.ctx = mock.Mock(jsonfile="file1", task_list={"process_name": "xin", "age": 38})
        mock_open_write = mock.mock_open()
        file_handle = mock_open_write.return_value.__enter__.return_value
        mock_open_file.return_value = mock_open_write.return_value
        mock_dumps.return_value = "data"

        utils.json_dumps()

        mock_open_file.assert_called_once_with("file1", "w")
        mock_dumps.assert_called_once_with(main.ctx.task_list, indent=2)
        file_handle.write.assert_called_once_with("data")
        file_handle.flush.assert_called_once_with()
        mock_fsync.assert_called_once_with(file_handle)

    @mock.patch('crmsh.crash_test.utils.crmshutils.this_node')
    @mock.patch('crmsh.crash_test.utils.msg_error')
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    def test_this_node_false(self, mock_run, mock_error, mock_this_node):
        mock_run.return_value = (1, None, "error data")
        mock_this_node.return_value = "node1"

        res = utils.this_node()
        self.assertEqual(res, "node1")

        mock_run.assert_called_once_with("crm_node --name")
        mock_error.assert_called_once_with("error data")
        mock_this_node.assert_called_once_with()
    
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    def test_this_node(self, mock_run):
        mock_run.return_value = (0, "data", None)
        res = utils.this_node()
        self.assertEqual(res, "data")
        mock_run.assert_called_once_with("crm_node --name")

    @mock.patch('crmsh.crash_test.utils.datetime')
    def test_str_to_datetime(self, mock_datetime):
        utils.str_to_datetime("Mon Nov  2 15:37:11 2020", "%a %b %d %H:%M:%S %Y")
        mock_datetime.strptime.assert_called_once_with("Mon Nov  2 15:37:11 2020", "%a %b %d %H:%M:%S %Y")

    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    def test_corosync_port_list(self, mock_run):
        output = """
totem.interface.0.bindnetaddr (str) = 10.10.10.121
totem.interface.0.mcastaddr (str) = 239.101.40.63
totem.interface.0.mcastport (u16) = 5405
totem.interface.0.ttl (u8) = 1
totem.interface.1.bindnetaddr (str) = 20.20.20.121
totem.interface.1.mcastaddr (str) = 239.6.213.31
totem.interface.1.mcastport (u16) = 5407
totem.interface.1.ttl (u8) = 1
        """
        mock_run.return_value = (0, output, None)
        result = utils.corosync_port_list()
        expected = ['5405', '5407']
        self.assertListEqual(result, expected)
        mock_run.assert_called_once_with("corosync-cmapctl totem.interface")

    def test_get_handler(self):
        mock_handler1 = mock.Mock(_name="test1_handler")
        mock_handler2 = mock.Mock(_name="test2_handler")
        mock_logger = mock.Mock(handlers=[mock_handler1, mock_handler2])
        res = utils.get_handler(mock_logger, "test1_handler")
        self.assertEqual(res, mock_handler1)

    @mock.patch('os.getuid')
    def test_is_root(self, mock_getuid):
        mock_getuid.return_value = 0
        self.assertEqual(utils.is_root(), True)
        mock_getuid.assert_called_once_with()

    @mock.patch('crmsh.crash_test.utils.crmshutils.to_ascii')
    @mock.patch('os.path.basename')
    @mock.patch('builtins.open')
    @mock.patch('os.path.join')
    @mock.patch('os.listdir')
    def test_get_process_status_false(self, mock_listdir, mock_join, mock_open_file, mock_basename, mock_to_ascii):
        mock_listdir.return_value = ['1', '2', 'none']
        mock_join.side_effect = ['/proc/1/cmdline', '/proc/2/cmdline']
        mock_open_read_1 = mock.mock_open(read_data=b'/usr/sbin/cmd1\x00--user\x00')
        mock_open_read_2 = mock.mock_open(read_data=b'/usr/sbin/cmd2\x00')
        mock_open_file.side_effect = [
            mock_open_read_1.return_value,
            mock_open_read_2.return_value
        ]
        mock_to_ascii.side_effect = [
            "/usr/sbin/cmd1\x00--user\x00",
            "/usr/sbin/cmd2\x00"
        ]
        mock_basename.side_effect = ["cmd1", "cmd2"]

        rc, pid = utils.get_process_status("sbd")
        self.assertEqual(rc, False)
        self.assertEqual(pid, -1)

        mock_listdir.assert_called_once_with('/proc')
        mock_join.assert_has_calls([
            mock.call('/proc', '1', 'cmdline'),
            mock.call('/proc', '2', 'cmdline')
            ])
        mock_open_file.assert_has_calls([
            mock.call('/proc/1/cmdline', 'rb'),
            mock.call('/proc/2/cmdline', 'rb')
            ])
        mock_to_ascii.assert_has_calls([
            mock.call(b'/usr/sbin/cmd1\x00--user\x00'),
            mock.call(b'/usr/sbin/cmd2\x00')
            ])

    @mock.patch('crmsh.crash_test.utils.crmshutils.to_ascii')
    @mock.patch('os.path.basename')
    @mock.patch('builtins.open')
    @mock.patch('os.path.join')
    @mock.patch('os.listdir')
    def test_get_process_status(self, mock_listdir, mock_join, mock_open_file, mock_basename, mock_to_ascii):
        mock_listdir.return_value = ['1', '2', 'none']
        mock_join.side_effect = ['/proc/1/cmdline', '/proc/2/cmdline']
        mock_open_read_1 = mock.mock_open(read_data=b'/usr/sbin/cmd1\x00--user\x00')
        mock_open_read_2 = mock.mock_open(read_data=b'/usr/sbin/sbd\x00')
        mock_open_file.side_effect = [
            mock_open_read_1.return_value,
            mock_open_read_2.return_value
        ]
        mock_to_ascii.side_effect = [
            "/usr/sbin/cmd1\x00--user\x00",
            "/usr/sbin/sbd\x00"
        ]
        mock_basename.side_effect = ["cmd1", "sbd"]

        rc, pid = utils.get_process_status("sbd")
        self.assertEqual(rc, True)
        self.assertEqual(pid, 2)

        mock_listdir.assert_called_once_with('/proc')
        mock_join.assert_has_calls([
            mock.call('/proc', '1', 'cmdline'),
            mock.call('/proc', '2', 'cmdline')
            ])
        mock_open_file.assert_has_calls([
            mock.call('/proc/1/cmdline', 'rb'),
            mock.call('/proc/2/cmdline', 'rb')
            ])
        mock_to_ascii.assert_has_calls([
            mock.call(b'/usr/sbin/cmd1\x00--user\x00'),
            mock.call(b'/usr/sbin/sbd\x00')
            ])

    @mock.patch('crmsh.crash_test.utils.msg_error')
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    def test_check_node_status_error_cmd(self, mock_run, mock_error):
        mock_run.return_value = (1, None, "error")
        res = utils.check_node_status("node1", "member")
        self.assertEqual(res, False)
        mock_run.assert_called_once_with("crm_node -l")
        mock_error.assert_called_once_with("error")

    @mock.patch('crmsh.crash_test.utils.msg_error')
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    def test_check_node_status(self, mock_run, mock_error):
        output = """
1084783297 15sp2-1 member
1084783193 15sp2-2 lost
        """
        mock_run.return_value = (0, output, None)

        res = utils.check_node_status("15sp2-2", "member")
        self.assertEqual(res, False)
        res = utils.check_node_status("15sp2-1", "member")
        self.assertEqual(res, True)

        mock_run.assert_has_calls([
            mock.call("crm_node -l"),
            mock.call("crm_node -l")
            ])
        mock_error.assert_not_called()

    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    def test_online_nodes_empty(self, mock_run):
        mock_run.return_value = (0, "data", None)
        res = utils.online_nodes()
        self.assertEqual(res, [])
        mock_run.assert_called_once_with("crm_mon -1")

    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    def test_online_nodes(self, mock_run):
        output = """
Node List:
  * Online: [ 15sp2-1 15sp2-2 ]
        """
        mock_run.return_value = (0, output, None)
        res = utils.online_nodes()
        self.assertEqual(res, ["15sp2-1", "15sp2-2"])
        mock_run.assert_called_once_with("crm_mon -1")

    @mock.patch('crmsh.crash_test.utils.online_nodes')
    def test_peer_node_list_empty(self, mock_online):
        mock_online.return_value = None
        res = utils.peer_node_list()
        self.assertEqual(res, [])
        mock_online.assert_called_once_with()

    @mock.patch('crmsh.crash_test.utils.this_node')
    @mock.patch('crmsh.crash_test.utils.online_nodes')
    def test_peer_node_list(self, mock_online, mock_this_node):
        mock_online.return_value = ["node1", "node2"]
        mock_this_node.return_value = "node1"
        res = utils.peer_node_list()
        self.assertEqual(res, ["node2"])
        mock_online.assert_called_once_with()

    # Test is_valid_sbd():
    @classmethod
    @mock.patch('os.path.exists')
    def test_is_valid_sbd_not_exist(cls, mock_os_path_exists):
        """
        Test device not exist
        """
        dev = "/dev/disk/by-id/scsi-device1"
        mock_os_path_exists.return_value = False

        res = utils.is_valid_sbd(dev)
        assert res is False

    @classmethod
    @mock.patch('crmsh.crash_test.utils.msg_error')
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('os.path.exists')
    def test_is_valid_sbd_cmd_error(cls, mock_os_path_exists,
                                    mock_sbd_check_header, mock_msg_err):
        """
        Test device is not valid sbd
        """
        dev = "/dev/disk/by-id/scsi-device1"
        mock_os_path_exists.return_value = True
        mock_sbd_check_header.return_value = (-1, None, "Unknown error!")
        mock_msg_err.return_value = ""

        res = utils.is_valid_sbd(dev)
        mock_msg_err.assert_called_once_with("Unknown error!")
        assert res is False

    @classmethod
    @mock.patch('crmsh.crash_test.utils.msg_error')
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('os.path.exists')
    def test_is_valid_sbd_not_sbd(cls, mock_os_path_exists,
                                  mock_sbd_check_header, mock_msg_err):
        """
        Test device is not SBD device
        """
        dev = "/dev/disk/by-id/scsi-device1"
        err_output = """
==Dumping header on disk {}
==Header on disk {} NOT dumped
sbd failed; please check the logs.
""".format(dev, dev)
        mock_os_path_exists.return_value = True
        mock_sbd_check_header.return_value = (1, "==Dumping header on disk {}".format(dev),
                                              err_output)

        res = utils.is_valid_sbd(dev)
        assert res is False
        mock_msg_err.assert_called_once_with(err_output)

    @classmethod
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('os.path.exists')
    def test_is_valid_sbd_is_sbd(cls, mock_os_path_exists,
                                 mock_sbd_check_header):
        """
        Test device is not SBD device
        """
        dev = "/dev/disk/by-id/scsi-device1"
        std_output = """
==Dumping header on disk {}
Header version     : 2.1
UUID               : f4c99362-6522-46fc-8ce4-7db60aff19bb
Number of slots    : 255
Sector size        : 512
Timeout (watchdog) : 5
Timeout (allocate) : 2
Timeout (loop)     : 1
Timeout (msgwait)  : 10
==Header on disk {} is dumped
""".format(dev, dev)
        mock_os_path_exists.return_value = True
        mock_sbd_check_header.return_value = (0, std_output, None)

        res = utils.is_valid_sbd(dev)
        assert res is True

    # Test find_candidate_sbd() and _find_match_count()
    @classmethod
    @mock.patch('glob.glob')
    @mock.patch('os.path.basename')
    @mock.patch('os.path.dirname')
    def test_find_candidate_no_dev(cls, mock_os_path_dname, mock_os_path_bname,
                                   mock_glob):
        """
        Test no suitable device
        """
        mock_os_path_dname.return_value = "/dev/disk/by-id"
        mock_os_path_bname.return_value = "scsi-label_CN_devA"
        mock_glob.return_value = []

        res = utils.find_candidate_sbd("/not-exist-folder/not-exist-dev")
        assert res == ""

    @classmethod
    @mock.patch('crmsh.crash_test.utils.is_valid_sbd')
    @mock.patch('glob.glob')
    @mock.patch('os.path.basename')
    @mock.patch('os.path.dirname')
    def test_find_candidate_no_can(cls, mock_os_path_dname, mock_os_path_bname,
                                   mock_glob, mock_is_valid_sbd):
        """
        Test no valid candidate device
        """
        mock_os_path_dname.return_value = "/dev/disk/by-id"
        mock_os_path_bname.return_value = "scsi-label_CN_devA"
        mock_glob.return_value = ["/dev/disk/by-id/scsi-label_DE_devA",
                                  "/dev/disk/by-id/scsi-label_DE_devB",
                                  "/dev/disk/by-id/scsi-label_DE_devC",
                                  "/dev/disk/by-id/scsi-label_DE_devD"]
        mock_is_valid_sbd.side_effect = [False, False, False, False]

        res = utils.find_candidate_sbd("/dev/disk/by-id/scsi-label_CN_devA")
        assert res == ""

    @classmethod
    @mock.patch('crmsh.crash_test.utils.is_valid_sbd')
    @mock.patch('glob.glob')
    @mock.patch('os.path.basename')
    @mock.patch('os.path.dirname')
    def test_find_candidate_has_multi(cls, mock_os_path_dname, mock_os_path_bname,
                                      mock_glob, mock_is_valid_sbd):
        """
        Test has multiple valid candidate devices
        """
        mock_os_path_dname.return_value = "/dev/disk/by-id"
        mock_os_path_bname.return_value = "scsi-label_CN_devA"
        mock_glob.return_value = ["/dev/disk/by-id/scsi-label_DE_devA",
                                  "/dev/disk/by-id/scsi-label_DE_devB",
                                  "/dev/disk/by-id/scsi-label_CN_devC",
                                  "/dev/disk/by-id/scsi-label_CN_devD",
                                  "/dev/disk/by-id/scsi-mp_China_devE",
                                  "/dev/disk/by-id/scsi-mp_China_devF"]
        mock_is_valid_sbd.side_effect = [True, False, False, True, True, False]

        res = utils.find_candidate_sbd("/dev/disk/by-id/scsi-label_CN_devA")
        assert res == "/dev/disk/by-id/scsi-label_CN_devD"
